import math
import sys
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Dict, List, Optional
import numpy as np
from dataclasses import dataclass, field
from contextlib import contextmanager

from negmas.java import to_java
from .common import ManufacturingProfile, Job, Factory, NO_PRODUCTION

__all__ = [
    'FactorySimulator',
    'SlowFactorySimulator',
    'FastFactorySimulator',
    'storage_as_array',
    'transaction',
    'temporary_transaction'
]


def storage_as_array(storage: Dict[int, int], n_products: int) -> np.array:
    """
    Converts storage to an array
    Args:
        storage: A dictionary giving quantity for each product index
        n_products: number of products (size of the resulting array)

    Returns:

    """
    a = np.zeros(n_products)
    for k, v in storage.items():
        a[k] = v
    return a


class FactorySimulator(ABC):
    """Simulates a factory allowing for prediction of storage/balance in the future"""

    def __init__(self, initial_wallet: float, initial_storage: Dict[int, int], n_steps: int, n_products: int
                 , profiles: List[ManufacturingProfile], max_storage: Optional[int] = None):
        self._n_steps = n_steps
        self._max_storage = max_storage if max_storage is not None else sys.maxsize
        self._initial_wallet = initial_wallet
        self._initial_storage = np.zeros(n_products)
        for k, v in initial_storage.items():
            self._initial_storage[k] = v
        self._profiles = profiles
        self._n_products = n_products
        self._reserved_storage = np.zeros(shape=(n_products, n_steps))

    def _as_array(self, storage: Dict[int, int]):
        return storage_as_array(storage=storage, n_products=self._n_products)

    # -----------------
    # FIXED PROPERTIES
    # -----------------

    @property
    def max_storage(self) -> Optional[int]:
        return self._max_storage

    @property
    def n_steps(self) -> int:
        return self._n_steps

    @property
    def initial_wallet(self) -> float:
        return self._initial_wallet

    @property
    def initial_storage(self) -> np.array:
        return self._initial_storage

    @property
    @abstractmethod
    def n_lines(self):
        """Number of lines"""

    @property
    @abstractmethod
    def final_balance(self) -> float:
        """Final balance given everything scheduled so-far"""

    # -------------------------------
    # DYNAMIC PROPERTIES (READ STATE)
    # -------------------------------

    @abstractmethod
    def wallet_to(self, t: int) -> np.array:
        """
        Returns the cash in wallet up to and including time t
        Args:
            t:

        Returns:

        """

    def wallet_at(self, t: int) -> float:
        return self.wallet_to(t)[-1]

    @abstractmethod
    def storage_to(self, t: int) -> np.array:
        """
        Returns the storage at time t

        Args:
            t:

        Returns:

        """

    def storage_at(self, t: int) -> np.array:
        return self.storage_to(t)[:, -1]

    @abstractmethod
    def line_schedules_to(self, t: int) -> np.array:
        """
        Returns the schedule of each line

        Args:
            t: time

        Returns:
            - A `NO_PRODUCTION` value means no production, otherwise the index of the process being run
        """

    def line_schedules_at(self, t: int) -> np.array:
        return self.line_schedules_to(t)[:, -1]

    def total_storage_to(self, t: int) -> np.array:
        return self.storage_to(t).sum(axis=0)

    def total_storage_at(self, t: int) -> int:
        return self.total_storage_to(t)[-1]

    def reserved_storage_to(self, t: int) -> np.array:
        return self._reserved_storage[:, :t + 1]

    def reserved_storage_at(self, t: int) -> np.array:
        return self._reserved_storage[:, t]

    def available_storage_at(self, t: int) -> np.array:
        return self.storage_at(t) - self.reserved_storage_at(t)

    def available_storage_to(self, t: int) -> np.array:
        return self.storage_to(t) - self.reserved_storage_to(t)

    @abstractmethod
    def loans_to(self, t: int) -> np.array:
        """
        Returns loans up to time t

        Args:
            t: time

        Returns:

        """

    def loans_at(self, t: int) -> float:
        """
        Returns loans at time t
        Args:
            t: time

        Returns:

        """
        return self.loans_to(t)[-1]

    def balance_at(self, t: int) -> float:
        """
        Returns the balance fo the factory at time t
        Args:
            t: time

        Returns:

        """
        return self.wallet_at(t) - self.loans_at(t)

    def balance_to(self, t: int) -> np.array:
        """
        Returns the balance fo the factory at times <= t
        Args:
            t: time

        Returns:

        """
        return self.wallet_to(t) - self.loans_to(t)

    @property
    @abstractmethod
    def fixed_before(self):
        """Gives the time before which the schedule is fixed"""

    # -------------------------
    # OPERATIONS (UPDATE STATE)
    # -------------------------

    @abstractmethod
    def set_state(self, t: int, storage: np.array, wallet: float, loans: float, line_schedules: np.array) -> None:
        """Sets the current state at the given time-step. It implicitly causes a fix_before(t + 1)"""

    @abstractmethod
    def add_loan(self, total: float, t: int) -> bool:
        """
        Adds a loan at the given time
        Args:
            total:
            t: time

        Returns:

        """

    def receive(self, payment: float, t: int) -> bool:
        """
        Simulates receiving payment at time t
        Args:
            payment:
            t: time

        Returns:

        """
        return self.pay(-payment, t)

    @abstractmethod
    def pay(self, payment: float, t: int, ignore_money_shortage: bool = True) -> bool:
        """
        Simulate payment at time t

        Args:
            payment:
            t: time
            ignore_money_shortage: If True, shortage in money will be ignored and the wallet can go negative

        Returns:

        """

    @abstractmethod
    def transport_to(self, product: int, quantity: int, t: int
                     , ignore_inventory_shortage: bool = True, ignore_space_shortage: bool = True) -> bool:
        """
        Simulates transporting products to/from storage at time t

        Args:
            product:
            quantity:
            t: time
            ignore_inventory_shortage: Ignore shortage in the `product` which may lead to negative storage[product]
            ignore_space_shortage:  Ignore the limit on total storage which may lead to total_storage > max_storage

        Returns:

        """

    @abstractmethod
    def buy(self, product: int, quantity: int, price: int, t: int
            , ignore_money_shortage: bool = True, ignore_space_shortage: bool = True) -> bool:
        """
        Buy a given quantity of a product for a given price at some time t

        Args:
            product:
            quantity:
            price:
            t: time
            ignore_money_shortage: If True, shortage in money will be ignored and the wallet can go negative
            ignore_space_shortage:  Ignore the limit on total storage which may lead to total_storage > max_storage

        Returns:

            - buy cannot ever have inventory shortage

        """

    @abstractmethod
    def sell(self, product: int, quantity: int, price: int, t: int
             , ignore_money_shortage: bool = True, ignore_inventory_shortage: bool = True) -> bool:
        """
        sell a given quantity of a product for a given price at some time t

        Args:
            product:
            quantity:
            price:
            t: time
            ignore_money_shortage: If True, shortage in money will be ignored and the wallet can go negative
            ignore_inventory_shortage: Ignore shortage in the `product` which may lead to negative storage[product]

        Returns:

        """

    @abstractmethod
    def schedule(self, job: Job, ignore_inventory_shortage=True, ignore_money_shortage=True, ignore_space_shortage=True
                 , override=True) -> bool:
        """
        Simulates scheduling the given job at its `time` and `line` optionally overriding whatever was already scheduled
        Args:
            job:
            ignore_inventory_shortage: If true shortages in inputs will be ignored
            ignore_money_shortage: If true, shortage in money will be ignored
            ignore_space_shortage: If true, shortage in space will be ignored
            override:

        Returns:
            Success/failure
        """

    def reserve(self, product: int, quantity: int, t: int) -> bool:
        """
        Simulates reserving the given quantity of the given product at times >= t
        Args:
            product:
            quantity:
            t: time

        Returns:

        """
        self._reserved_storage[product, t] += quantity
        return True

    # ------------------
    # HISTORY MANAGEMENT
    # ------------------

    @abstractmethod
    def fix_before(self, t: int) -> bool:
        """
        Fix the history before this point

        Args:
            t: time

        Returns:

        """

    @abstractmethod
    def bookmark(self) -> int:
        """Sets a bookmark to the current location

        Returns:
            bookmark ID
        """

    @abstractmethod
    def rollback(self, bookmark_id: int) -> bool:
        """Rolls back to the given bookmark ID

        Args:
            bookmark_id The bookmark ID returned from bookmark

        Remarks:

            - You can only rollback in the reverse order of bookmarks. If the bookmark ID given here is not the one
              at the top of the bookmarks stack, the rollback will fail (return False)

        """

    @abstractmethod
    def delete_bookmark(self, bookmark_id: int) -> bool:
        """Commits everything since the bookmark so it cannot be rolled back

        Args:
            bookmark_id The bookmark ID returned from bookmark

        Remarks:

            - You can only rollback in the reverse order of bookmarks. If the bookmark ID given here is not the one
              at the top of the bookmarks stack, the deletion will fail (return False)

        """


@dataclass
class _Bookmark:
    id: int
    jobs: Dict[int, List[int]] = field(default_factory=lambda: defaultdict(list), init=False)
    buy_contracts: Dict[int, List[int]] = field(default_factory=lambda: defaultdict(list), init=False)
    sell_contracts: Dict[int, List[int]] = field(default_factory=lambda: defaultdict(list), init=False)
    payment_updates: Dict[int, float] = field(default_factory=lambda: defaultdict(float), init=False)
    loans_updates: Dict[int, float] = field(default_factory=lambda: defaultdict(float), init=False)
    storage_updates: Dict[int, Dict[int, int]] = field(default_factory=lambda: defaultdict(lambda: defaultdict(int))
                                                       , init=False)


@dataclass
class _State:
    t: int
    storage: np.array
    wallet: float
    loans: float
    line_schedules: np.array


class SlowFactorySimulator(FactorySimulator):
    """A slow factory simulator that runs an internal factory to find-out what will happen in the future

    Remarks:

        - It is *much* faster to always access the properties/methods of this class in ascending time. If that is not
          the case, each time reversal will cause a complete reset.
        - It is recommended to call `fix_before` () to fix the past once a production step is completed. That will speed
          up operations
    """

    def set_state(self, t: int, storage: np.array, wallet: float, loans: float, line_schedules: np.array) -> None:
        for i, s in enumerate(storage):
            d = s - self.storage_at(t)[i]
            if d == 0.0:
                continue
            self._storage_updates[t][i] += d
        d = self.wallet_at(t) - wallet
        if d != 0.0:
            self._payment_updates[t] -= d
        d = self.loans_at(t) - loans
        if d != 0.0:
            self._loans_updates[t] -= d
        expected_schedules = self.line_schedules_at(t)
        for i in range(self.n_lines):
            expected, actual = expected_schedules[i], line_schedules[i]
            if expected == actual:
                continue
            if expected == NO_PRODUCTION:
                raise ValueError(f'Expected no production at time {t} on line {i} but actually process '
                                 f'{actual} is running')
            if expected != actual and actual != NO_PRODUCTION:
                raise ValueError(f'Expected process {expected} at time {t} on line {i} but actually process '
                                 f'{actual} is running')
            self._line_schedules[i, t] = actual
        self.fix_before(t + 1)
        self._saved_states[t].append(_State(t=t, storage=storage.copy(), wallet=wallet, loans=loans
                                            , line_schedules=line_schedules.copy()))

    def delete_bookmark(self, bookmark_id: int) -> bool:
        if self._active_bookmark is None or self._active_bookmark.id != bookmark_id:
            raise ValueError(f'there is no active bookmark to delete')
        self._bookmarks, self._bookmarked_at = self._bookmarks[:-1], self._bookmarked_at[:-1]
        self._active_bookmark = self._bookmarks[-1] if len(self._bookmarks) > 0 else None
        self._active_bookmarked_at = self._bookmarked_at[-1] if len(self._bookmarked_at) > 0 else -1
        return True

    def bookmark(self) -> int:
        bookmark = _Bookmark(id=len(self._bookmarks))
        self._bookmarks.append(bookmark)
        self._bookmarked_at.append(self._factory.next_step)
        self._active_bookmark = bookmark
        self._active_bookmarked_at = self._bookmarked_at[-1]
        return bookmark.id

    def rollback(self, bookmark_id: int) -> bool:
        if self._active_bookmark is None or self._active_bookmark.id != bookmark_id:
            raise ValueError(f'there is no active bookmark to rollback')
        for t, payment in self._active_bookmark.payment_updates.items():
            self._payment_updates[t] += payment
        for t, payment in self._active_bookmark.loans_updates.items():
            self._loans_updates[t] += payment
        for t, storage in self._active_bookmark.storage_updates.items():
            s = self._storage_updates[t]
            for k, v in storage:
                s[k] -= v
        for t, rolled_indices in self._active_bookmark.jobs.items():
            self._jobs[t] = [_ for i, _ in enumerate(self._jobs[t]) if i not in rolled_indices]
        for t, rolled_indices in self._active_bookmark.buy_contracts.items():
            self._buy_contracts[t] = [_ for i, _ in enumerate(self._buy_contracts[t]) if i not in rolled_indices]
        for t, rolled_indices in self._active_bookmark.sell_contracts.items():
            self._sell_contracts[t] = [_ for i, _ in enumerate(self._sell_contracts[t]) if i not in rolled_indices]

        if self._factory.next_step != self._bookmarked_at:
            self.goto(self._active_bookmarked_at)
        return True

    @property
    def final_balance(self) -> float:
        self.goto(self.n_steps - 1)
        return self.balance_at(self.n_steps - 1)

    @property
    def n_lines(self):
        return self._factory.n_lines

    def fix_before(self, t: int) -> bool:
        self.goto(t)
        self._fixed_before = t
        invalid = [i for i, bt in enumerate(self._bookmarked_at) if bt < t]
        self._bookmarks = [_ for i, _ in enumerate(self._bookmarks) if i not in invalid]
        self._bookmarked_at = [_ for i, _ in enumerate(self._bookmarked_at) if i not in invalid]
        return True

    def __init__(self, initial_wallet: float, initial_storage: Dict[int, int], n_steps: int, n_products: int
                 , profiles: List[ManufacturingProfile], max_storage: Optional[int]):
        super().__init__(initial_wallet=initial_wallet, initial_storage=initial_storage, n_steps=n_steps
                         , n_products=n_products, profiles=profiles, max_storage=max_storage)
        self._factory = Factory(initial_storage=initial_storage, initial_wallet=initial_wallet
                                , profiles=profiles, max_storage=max_storage)
        self._jobs: Dict[int, List[(Job, bool, bool, bool, bool)]] = defaultdict(list)
        self._buy_contracts: Dict[int, List[(int, int, float)]] = defaultdict(list)
        self._sell_contracts: Dict[int, List[(int, int, float)]] = defaultdict(list)
        self._payment_updates: Dict[int, float] = defaultdict(float)
        self._loans_updates: Dict[int, float] = defaultdict(float)
        self._storage_updates: Dict[int, Dict[int, int]] = defaultdict(lambda: defaultdict(int))
        self._wallet = np.zeros(n_steps)
        self._loans = np.zeros(n_steps)
        self._storage = np.zeros(shape=(n_products, n_steps))
        self._line_schedules = np.zeros(shape=(self._factory.n_lines, self._n_steps))
        self._fixed_before = 0
        self._bookmarks: List[_Bookmark] = []
        self._active_bookmark: Optional[_Bookmark] = None
        self._active_bookmarked_at: int = -1
        self._bookmarked_at: List[int] = []
        self._saved_states: Dict[int, List[_State]] = defaultdict(list)

    def _update_state(self) -> None:
        t = self._factory.next_step - 1
        if t < 0:
            return
        self._wallet[t] = self._factory.wallet
        self._loans[t] = self._factory.loans
        self._storage[:, t] = self._as_array(self._factory.storage)
        self._line_schedules[:, t] = np.array(list(NO_PRODUCTION if command.is_none else command.profile.process.id
                                                   for command in self._factory.commands))

    def reset_to(self, t: int) -> None:
        self._factory = Factory(initial_storage={i: v for i, v in enumerate(self._initial_storage) if v != 0}
                                , initial_wallet=self._initial_wallet
                                , profiles=self._profiles)
        for step in range(t + 1):
            self._factory.receive(payment=self._payment_updates.get(step, 0.0))
            self._factory.add_loan(total=self._loans_updates.get(step, 0.0))
            jobs = self._jobs.get(t, [])
            for job, override, ignore_storage, ignore_money, ignore_space in jobs:
                # @todo use ignore* here
                try:
                    self._factory.schedule(job=job, override=override)
                except ValueError as err:
                    print(err)
            contracts = self._buy_contracts.get(t, [])
            for product, quantity, price in contracts:
                try:
                    self._factory.buy(product=product, quantity=quantity, price=price)
                except ValueError as err:
                    print(err)
            contracts = self._sell_contracts.get(t, [])
            for product, quantity, price in contracts:
                try:
                    self._factory.sell(product=product, quantity=quantity, price=price)
                except ValueError as err:
                    print(err)
            inventory = self._storage_updates.get(step, {})
            for product, quantity in inventory.items():
                try:
                    self._factory.transport_to(product, quantity)
                except ValueError as err:
                    print(err)
            self._update_state()

    def goto(self, t: int) -> None:
        """
        Steps the factory to the end of step t
        Args:
            t: time

        Returns:

        """
        if t > self.n_steps - 1:
            t = self.n_steps - 1
        if self._factory.next_step > t + 1:
            if t < self._fixed_before:
                return
            self.reset_to(t)
        while self._factory.next_step <= t:
            step = self._factory.next_step
            loan = self._loans_updates.get(step, None)
            if loan is not None:
                self._factory.add_loan(loan)
            payment = self._payment_updates.get(step, None)
            if payment is not None:
                self._factory.pay(payment)
            jobs = self._jobs.get(step, [])
            for job, override, ignore_storage, ignore_money, ignore_space in jobs:
                # @todo implement ignore_input_shortage inside the factory to use ignore_input_shortage here
                self._factory.schedule(job=job, override=override)
            self._factory.step()
            inventory = self._storage_updates.get(step, None)
            if inventory is not None:
                for product, quantity in inventory.items():
                    self._factory.transport_to(product, quantity)
            self._update_state()

    def wallet_to(self, t: int) -> np.array:
        if t < self._fixed_before:
            return self._wallet[:t + 1]
        self.goto(t)
        return self._wallet[:t + 1]

    def line_schedules_to(self, t: int) -> np.array:
        if t < self._fixed_before:
            return self._storage[:, :t + 1]
        self.goto(t)
        return self._line_schedules[:, :t + 1]

    def storage_to(self, t: int) -> np.array:
        if t < self._fixed_before:
            return self._storage[:, :t + 1]
        self.goto(t)
        return self._storage[:, :t + 1]

    def loans_to(self, t: int) -> float:
        if t < self._fixed_before:
            return self._loans[:t + 1]
        self.goto(t)
        return self._loans[:t + 1]

    def add_loan(self, total: float, t: int) -> bool:
        if t < self._fixed_before:
            raise ValueError(f'Cannot run operations in the past (t={t}, fixed before {self._fixed_before})')
        self._loans_updates[t] += total
        if self._active_bookmark:
            self._active_bookmark.loans_updates[t] += total
        return True

    def pay(self, payment: float, t: int, ignore_money_shortage: bool = True) -> bool:
        if t < self._fixed_before:
            raise ValueError(f'Cannot run operations in the past (t={t}, fixed before {self._fixed_before})')
        self._payment_updates[t] += payment
        if self._active_bookmark:
            self._active_bookmark.payment_updates[t] += payment
        return True

    def transport_to(self, product: int, quantity: int, t: int, ignore_inventory_shortage: bool = True, ignore_space_shortage: bool = True) -> bool:
        if t < self._fixed_before:
            raise ValueError(f'Cannot run operations in the past (t={t}, fixed before {self._fixed_before})')
        s = self._storage_updates[t]
        s[product] += quantity
        if self._active_bookmark:
            s = self._active_bookmark.storage_updates[t]
            s[product] += quantity
        return True

    def schedule(self, job: Job, ignore_inventory_shortage=True, ignore_money_shortage=True, ignore_space_shortage=True
                 , override=True) -> bool:
        t = job.time
        if t < self._fixed_before:
            raise ValueError(f'Cannot run operations in the past (t={t}, fixed before {self._fixed_before})')
        self._jobs[t].append((job, override, ignore_inventory_shortage, ignore_money_shortage, ignore_space_shortage))
        if self._active_bookmark:
            self._active_bookmark.jobs[t].append(len(self._jobs[t]))
        return True

    def buy(self, product: int, quantity: int, price: int, t: int
            , ignore_money_shortage: bool = True, ignore_space_shortage: bool = True) -> bool:
        if t < self._fixed_before:
            raise ValueError(f'Cannot run operations in the past (t={t}, fixed before {self._fixed_before})')
        self._buy_contracts[t].append((product, quantity, price))
        if self._active_bookmark:
            self._active_bookmark.buy_contracts[t].append(len(self._buy_contracts[t]))
        return True

    def sell(self, product: int, quantity: int, price: int, t: int, ignore_money_shortage: bool = True, ignore_inventory_shortage: bool = True) -> bool:
        if t < self._fixed_before:
            raise ValueError(f'Cannot run operations in the past (t={t}, fixed before {self._fixed_before})')
        self._sell_contracts[t].append((product, quantity, price))
        if self._active_bookmark:
            self._active_bookmark.sell_contracts[t].append(len(self._sell_contracts[t]))
        return True

    @property
    def fixed_before(self):
        return self._fixed_before


@dataclass
class _FullBookmark:
    id: int
    wallet: np.array
    loans: np.array
    storage: np.array
    line_schedules: np.array
    has_jobs: np.array


class FastFactorySimulator(FactorySimulator):

    def _as_array(self, storage: Dict[int, int]) -> np.array:
        a = np.zeros(self._n_products)
        for k, v in storage.items():
            a[k] = v
        return a

    def __init__(self, initial_wallet: float, initial_storage: Dict[int, int], n_steps: int, n_products: int
                 , profiles: List[ManufacturingProfile], max_storage: Optional[int]):
        super().__init__(initial_wallet=initial_wallet, initial_storage=initial_storage, n_steps=n_steps
                         , n_products=n_products, profiles=profiles, max_storage=max_storage)
        self._wallet = np.ones(n_steps) * initial_wallet
        self._loans = np.zeros(n_steps)
        self._storage = np.repeat(self._as_array(initial_storage).reshape((n_products, 1)), n_steps, axis=1)
        self._total_storage = self._storage.sum(axis=0)
        factory = Factory(initial_storage=initial_storage, initial_wallet=initial_wallet, profiles=profiles
                          , max_storage=max_storage)
        self._profiles = factory.profiles
        self._n_lines = factory.n_lines
        self._line_schedules = np.ones(shape=(self._n_lines, self._n_steps)) * NO_PRODUCTION
        self._has_jobs = np.zeros(shape=(self._n_lines, self._n_steps), dtype=bool)
        self._fixed_before = 0
        self._bookmarks: List[_FullBookmark] = []
        self._active_bookmark: Optional[_FullBookmark] = None

    def init(self, *args, **kwargs):
        self.__init__(*args, **kwargs)

    @property
    def fixed_before(self):
        return self._fixed_before

    @property
    def n_lines(self):
        return self._n_lines

    @property
    def final_balance(self) -> float:
        return self._wallet[-1] - self._loans[-1]

    def wallet_to(self, t: int) -> np.array:
        return self._wallet[:t + 1]

    def storage_to(self, t: int) -> np.array:
        return self._storage[:, :t + 1]

    def line_schedules_to(self, t: int) -> np.array:
        return self._line_schedules[:, : t + 1]

    def loans_to(self, t: int) -> np.array:
        return self._loans[: t + 1]

    def add_loan(self, total: float, t: int) -> bool:
        if t < self._fixed_before:
            raise ValueError(f'Cannot run operations in the past (t={t}, fixed before {self._fixed_before})')
        self._loans[t:] += total
        return True

    def pay(self, payment: float, t: int, ignore_money_shortage: bool = True) -> bool:
        # @todo add minimum balance
        if t < self._fixed_before:
            raise ValueError(f'Cannot run operations in the past (t={t}, fixed before {self._fixed_before})')
        b = self._wallet[t:]
        b -= payment
        if b.min() < 0:
            b += payment
            return False
        return True

    def transport_to(self, product: int, quantity: int, t: int, ignore_inventory_shortage: bool = True, ignore_space_shortage: bool = True) -> bool:
        # @todo add minimum storage
        if t < self._fixed_before:
            raise ValueError(f'Cannot run operations in the past (t={t}, fixed before {self._fixed_before})')
        s, total = self._storage[product, t:].view(), self._total_storage[t:]
        s += quantity
        total += quantity
        if s.min() < 0 or total.max() > self.max_storage:
            s -= quantity
            total -= quantity
            return False
        return True

    def buy(self, product: int, quantity: int, price: int, t: int
            , ignore_money_shortage: bool = True, ignore_space_shortage: bool = True) -> bool:
        if t < self._fixed_before:
            raise ValueError(f'Cannot run operations in the past (t={t}, fixed before {self._fixed_before})')
        s, total = self._storage[product, t:].view(), self._total_storage[t:]
        s += quantity
        total += quantity
        b = self._wallet[t:]
        b -= price
        if total.max() > self.max_storage or b.min() < 0:
            s -= quantity
            total -= quantity
            b += price
            return False
        return True

    def sell(self, product: int, quantity: int, price: int, t: int, ignore_money_shortage: bool = True, ignore_inventory_shortage: bool = True) -> bool:
        if t < self._fixed_before:
            raise ValueError(f'Cannot run operations in the past (t={t}, fixed before {self._fixed_before})')
        s, total = self._storage[product, t:].view(), self._total_storage[t:]
        s -= quantity
        total -= quantity
        b = self._wallet[t:]
        b += price
        if s.min() < 0:
            s += quantity
            total += quantity
            b -= price
            return False
        return True

    def schedule(self, job: Job, ignore_inventory_shortage=True, ignore_money_shortage=True, ignore_space_shortage=True
                 , override=True) -> bool:
        t, job_override = job.time, job.override
        if t < self._fixed_before:
            raise ValueError(f'Cannot run operations in the past (t={t}, fixed before {self._fixed_before})')
        if job_override:
            raise NotImplementedError(f'{self.__class__.__name__} does not support scheduling jobs with overriding')
        # job_line = job.line # only useful for stop/pause/resume that are not supported
        profile = self._profiles[job.profile]
        inputs, outputs, length, cost = profile.process.inputs, profile.process.outputs, profile.n_steps, profile.cost
        line = profile.line

        # confirm that there is no other jobs already scheduled at this exact time:
        if self._has_jobs[line, t]:
            if override:
                raise NotImplementedError(f'{self.__class__.__name__} does not support scheduling more than a single '
                                          f'job at any time-step/line')
            return False

        # confirm that the line is not busy. If it was busy, and we are not overriding, fail.
        if not job_override and np.any(self._line_schedules[line, t:t + length] != NO_PRODUCTION):
            return False

        # confirm that there is enough money to start production
        if (not ignore_money_shortage) and np.any(self._wallet[t:] < cost):
            return False
        # bookmark to be able to rollback at any error
        if job.action == 'run':
            with transaction(self) as bookmark:
                if not self.pay(cost, t):
                    self.rollback(bookmark)
                    return False
                self._line_schedules[line, t:t + length] = profile.process.id
                for i in inputs:
                    it = int(math.floor(i.step * length) + t)
                    p, q = i.product, i.quantity
                    if (not ignore_inventory_shortage) and np.any(self._storage[p, it:] < q):
                        self.rollback(bookmark)
                        return False
                    s, total = self._storage[p, it:].view(), self._total_storage[it:]
                    s -= q
                    total -= q
                for o in outputs:
                    ot = int(math.ceil(o.step * length) + t)
                    p, q = o.product, o.quantity
                    if (not ignore_space_shortage) and np.any(self._total_storage[ot:] + q > self.max_storage):
                        self.rollback(bookmark)
                        return False
                    s, total = self._storage[p, ot:].view(), self._total_storage[ot:]
                    s += q
                    total += q
            return True
        raise NotImplementedError(f'{self.__class__.__name__} does not support scheduling {job.action} jobs')

    def fix_before(self, t: int) -> bool:
        self._fixed_before = t
        return True

    def delete_bookmark(self, bookmark_id: int) -> bool:
        if self._active_bookmark is None or self._active_bookmark.id != bookmark_id:
            raise ValueError(f'there is no active bookmark to delete')
        self._bookmarks = self._bookmarks[:-1]
        self._active_bookmark = self._bookmarks[-1] if len(self._bookmarks) > 0 else None
        return True

    def bookmark(self) -> int:
        bookmark = _FullBookmark(id=len(self._bookmarks), wallet=self._wallet.copy(), loans=self._loans.copy()
                                 , storage=self._storage.copy(), line_schedules=self._line_schedules.copy()
                                 , has_jobs=self._has_jobs.copy())
        self._bookmarks.append(bookmark)
        self._active_bookmark = bookmark
        return bookmark.id

    def rollback(self, bookmark_id: int) -> bool:
        if self._active_bookmark is None or self._active_bookmark.id != bookmark_id:
            raise ValueError(f'there is no active bookmark to rollback')
        b = self._active_bookmark
        self._wallet, self._loans, self._storage = b.wallet, b.loans, b.storage
        self._line_schedules, self._has_jobs = b.line_schedules, b.has_jobs
        self._total_storage = self._storage.sum(axis=0)
        return True

    def set_state(self, t: int, storage: np.array, wallet: float, loans: float, line_schedules: np.array) -> None:
        self._storage[:, t:] += storage.reshape(self._n_products, 1) - self._storage[:, t].reshape(self._n_products, 1)
        self._wallet[t:] += wallet - self._wallet[t]
        self._loans[t:] += loans - self._loans[t]

        self._line_schedules[:, t] = line_schedules

        # @todo enable this again to confirm that simulation is correct. may be I set_state before the job is run on the simulator
        # expected_schedules = self._line_schedules[:, t]
        # for i in range(self.n_lines):
        #     expected, actual = expected_schedules[i], line_schedules[i]
        #     if expected == actual:
        #         continue
        #     if expected == NO_PRODUCTION:
        #         raise ValueError(f'Expected no production at time {t} on line {i} but actually process '
        #                          f'{actual} is running')
        #     if expected != actual and actual != NO_PRODUCTION:
        #         raise ValueError(f'Expected process {expected} at time {t} on line {i} but actually process '
        #                          f'{actual} is running')
        #     self._line_schedules[i, t] = actual

        self.fix_before(t)

    class Java:
        implements = ['jnegmas.apps.scml.simulators.FactorySimulator']


@contextmanager
def transaction(simulator):
    """Runs the simulated actions then confirms them if they are not rolled back"""
    _bookmark = simulator.bookmark()
    yield _bookmark
    simulator.delete_bookmark(_bookmark)


@contextmanager
def temporary_transaction(simulator):
    """Runs the simulated actions then rolls them back"""
    _bookmark = simulator.bookmark()
    yield _bookmark
    simulator.rollback(_bookmark)
    simulator.delete_bookmark(_bookmark)


class _ShadowFactorySimulator:
    """An FactorySimulator As seen by JNegMAS.

        This is an object that is not visible to python code. It is not directly called from python ever. It is only called
        from a corresponding Java object to represent an internal python object. Because of he way py4j works, we cannot
        just use dunders to implement this kind of object in general. We will have to implement each such class
        independently.

        This kind of classes will always have an internal Java class implementing a Java interface in Jnegmas that starts
        with Py.

    """

    def __init__(self, simulator: FactorySimulator):
        self.simulator = simulator

    def maxStorage(self) -> int:
        return to_java(self.simulator.max_storage)

    def nSteps(self) -> int:
        return to_java(self.simulator.n_steps)

    def initialWallet(self) -> float:
        return to_java(self.simulator.initial_wallet)

    def initialStorage(self) -> List[int]:
        return to_java(self.simulator.initial_storage.tolist())

    def nLines(self) -> int:
        return to_java(self.simulator.n_lines)

    def finalBalance(self) -> float:
        return to_java(self.simulator.final_balance)

    def walletTo(self, t: int) -> List[float]:
        return to_java(self.simulator.wallet_to(t).tolist())

    def walletAt(self, t: int) -> float:
        return to_java(self.simulator.wallet_at(t))

    def storageTo(self, t: int) -> List[int]:
        return to_java(self.simulator.storage_to(t).tolinst())

    def storageAt(self, t: int) -> int:
        return to_java(self.simulator.storage_at(t))

    def line_schedules_to(self, t: int) -> List[List[int]]:
        return to_java(self.simulator.line_schedules_to(t).tolist())

    def line_schedules_at(self, t: int) -> List[int]:
        return to_java(self.simulator.line_schedules_at(t).tolist())

    def totalStorageTo(self, t: int) -> List[int]:
        return to_java(self.simulator.total_storage_to(t).tolist())

    def totalStorageAt(self, t: int) -> int:
        return to_java(self.simulator.total_storage_at(t))

    def reservedStorageTo(self, t: int) -> List[int]:
        return to_java(self.simulator.reserved_storage_to(t).tolist())

    def reservedStorageAt(self, t: int) -> int:
        return to_java(self.simulator.reserved_storage_at(t))

    def availableStorageTo(self, t: int) -> List[int]:
        return to_java(self.simulator.available_storage_to(t).tolist())

    def availableStorageAt(self, t: int) -> int:
        return to_java(self.simulator.available_storage_at(t))

    def loansTo(self, t: int) -> List[float]:
        return to_java(self.simulator.loans_to(t).tolist())

    def loansAt(self, t: int) -> float:
        return to_java(self.simulator.loans_at(t))

    def balanceTo(self, t: int) -> List[float]:
        return to_java(self.simulator.balance_to(t).tolist())

    def balanceAt(self, t: int) -> float:
        return to_java(self.simulator.balance_at(t))

    def fixedBefore(self) -> int:
        return to_java(self.simulator.fixed_before)

    def setState(self, t: int, storage: List[int], wallet: float, loans: float, lineSchedules: List[int]) -> None:
        self.simulator.set_state(t, storage=storage, wallet=wallet, loans=loans
                                 , line_schedules=lineSchedules)

    def addLoan(self, total: float, t: int) -> bool:
        return self.simulator.add_loan(total=total, t=t)

    def receive(self, payment, t):
        return self.simulator.receive(payment, t)

    def pay(self, payment, t, ignoreMoneyShortage):
        return self.simulator.pay(payment, t, ignoreMoneyShortage)

    def transportTo(self, product, quantity, t, ignoreInventoryShortage, ignoreSpaceShortage):
        return self.simulator.transport_to(product, quantity, t, ignoreInventoryShortage, ignoreSpaceShortage)

    def buy(self, product, quantity, price, t, ignoreMoneyShortage, ignoreSpaceShortage):
        return self.simulator.buy(product, quantity, price, t, ignoreMoneyShortage, ignoreSpaceShortage)

    def sell(self, product, quantity, price, t, ignoreMoneyShortage, ignoreInventoryShortage):
        return self.simulator.sell(product, quantity, price, t, ignoreMoneyShortage, ignoreInventoryShortage)

    def schedule(self, job, ignoreInventoryShortage, ignoreMoneyShortage, ignoreSpaceShortage, override):
        return self.simulator.schedule(job, ignoreInventoryShortage, ignoreMoneyShortage, ignoreSpaceShortage, override)

    def reserve(self, product, quantity, t):
        return self.reserve(product, quantity, t)

    def fixBefore(self, t):
        return self.simulator.fix_before(t)

    def bookmark(self):
        return self.simulator.bookmark()

    def rollback(self, bookmarkId):
        return self.simulator.rollback(bookmarkId)

    def deleteBookmark(self, bookmarkId):
        return self.simulator.delete_bookmark(bookmarkId)

    class Java:
        implements = ['jengmas.apps.scml.simulators.FactorySimulator']




from os import walk
import os

import pkg_resources
import pytest

from negmas import load_genius_domain_from_folder, AspirationNegotiator


@pytest.fixture
def scenarios_folder():
    return pkg_resources.resource_filename('negmas', resource_name='tests/data/scenarios')


def test_importing_file_without_exceptions(scenarios_folder):
    folder_name = scenarios_folder + '/other/S-1NIKFRT-1'
    domain = load_genius_domain_from_folder(folder_name
                                            , n_discretization=10
                                            )
    # print(domain)


def test_convert_dir_keep_names(tmpdir):
    from negmas import convert_genius_domain_from_folder
    dst = tmpdir.mkdir("sub")
    src = pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop')
    dst = pkg_resources.resource_filename('negmas', resource_name='tests/data/LaptopConv')
    assert convert_genius_domain_from_folder(src_folder_name=src
                                      , dst_folder_name=dst
                                      , force_single_issue=True
                                      , cache_and_discretize_outcomes=True
                                      , n_discretization=10
                                      , keep_issue_names=True
                                      , keep_value_names=True
                                      , normalize_utilities=True)
    mechanism, agent_info, issues = load_genius_domain_from_folder(dst)
    assert len(issues) == 1
    for k, v in enumerate(issues):
        assert f'{k}:{v}' == '''0:Laptop-Harddisk-External Monitor: ["Dell+60 Gb+19'' LCD", "Dell+60 Gb+20'' LCD", "Dell+60 Gb+23'' LCD", "Dell+80 Gb+19'' LCD", "Dell+80 Gb+20'' LCD", "Dell+80 Gb+23'' LCD", "Dell+120 Gb+19'' LCD", "Dell+120 Gb+20'' LCD", "Dell+120 Gb+23'' LCD", "Macintosh+60 Gb+19'' LCD", "Macintosh+60 Gb+20'' LCD", "Macintosh+60 Gb+23'' LCD", "Macintosh+80 Gb+19'' LCD", "Macintosh+80 Gb+20'' LCD", "Macintosh+80 Gb+23'' LCD", "Macintosh+120 Gb+19'' LCD", "Macintosh+120 Gb+20'' LCD", "Macintosh+120 Gb+23'' LCD", "HP+60 Gb+19'' LCD", "HP+60 Gb+20'' LCD", "HP+60 Gb+23'' LCD", "HP+80 Gb+19'' LCD", "HP+80 Gb+20'' LCD", "HP+80 Gb+23'' LCD", "HP+120 Gb+19'' LCD", "HP+120 Gb+20'' LCD", "HP+120 Gb+23'' LCD"]'''


def test_convert_dir_no_names(tmpdir):
    from negmas import convert_genius_domain_from_folder
    dst = tmpdir.mkdir("sub")
    src = pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop')
    dst = pkg_resources.resource_filename('negmas', resource_name='tests/data/LaptopConv')

    assert convert_genius_domain_from_folder(src_folder_name=src
                                             , dst_folder_name=dst
                                             , force_single_issue=True
                                             , cache_and_discretize_outcomes=True
                                             , n_discretization=10
                                             , keep_issue_names=False
                                             , keep_value_names=False
                                             , normalize_utilities=True)
    mechanism, agent_info, issues = load_genius_domain_from_folder(dst)
    assert len(issues) == 1
    for k, v in enumerate(issues):
        assert f'{k}:{v}' == '''0:0: ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23', '24', '25', '26']'''


def test_simple_run_with_aspiration_agents():
    file_name = pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop')
    assert os.path.exists(file_name)
    mechanism, agents, issues = load_genius_domain_from_folder(
        file_name, n_steps=100, time_limit=30
        , force_single_issue=True, keep_issue_names=False
        , keep_value_names=False, agent_factories=AspirationNegotiator)
    assert mechanism is not None
    state = mechanism.run()




import math
import sys
from os import PathLike
from random import randint

from negmas.helpers import get_class
from negmas.tournaments import WorldRunResults, TournamentResults, tournament
from .factory_managers import GreedyFactoryManager
from .world import SCMLWorld

if True:
    from typing import Tuple, Union, Type, Iterable, Sequence, Optional, Callable, Any, Dict
    from .factory_managers import FactoryManager

__all__ = [
    'anac2019_world',
    'anac2019_tournament',
    'balance_calculator',
]


def anac2019_world(
    competitors: Sequence[Union[str, Type[FactoryManager]]] = ()
    , params: Sequence[Dict[str, Any]] = ()
    , randomize: bool = True
    , log_file_name: str = None
    , name: str = None
    , agent_names_reveal_type: bool = False
    , n_intermediate: Tuple[int, int] = (1, 4)
    , n_miners=5
    , n_factories_per_level=11
    , n_agents_per_competitor=1
    , n_consumers=5
    , n_lines_per_factory=10
    , guaranteed_contracts=False
    , use_consumer=True
    , max_insurance_premium=100, n_retrials=5
    , negotiator_type: str = 'negmas.sao.AspirationNegotiator'
    , transportation_delay=0
    , default_signing_delay=0
    , max_storage=sys.maxsize
    , consumption_horizon=15
    , consumption=(3, 5)
    , negotiation_speed=21
    , neg_time_limit=60 * 4
    , neg_n_steps=20
    , n_steps=100
    , time_limit=60 * 90
    , n_default_per_level: int = 5

) -> SCMLWorld:
    """
    Creates a world compatible with the ANAC 2019 competition. Note that

    Args:
        n_agents_per_competitor: Number of instantiations of each competing type.
        name: World name to use
        agent_names_reveal_type: If true, a snake_case version of the agent_type will prefix agent names
        randomize: If true, managers are assigned to factories randomly otherwise in the order
        they are giving (cycling).
        n_intermediate:
        n_default_per_level:
        competitors: A list of class names for the competitors
        params: A list of dictionaries giving parameters to pass to the competitors
        n_miners: number of miners of the single raw material
        n_factories_per_level: number of factories at every production level
        n_consumers: number of consumers of the final product
        n_steps: number of simulation steps
        n_lines_per_factory: number of lines in each factory
        negotiation_speed: The number of negotiation steps per simulation step. None means infinite
        default_signing_delay: The number of simulation between contract conclusion and signature
        neg_n_steps: The maximum number of steps of a single negotiation (that is double the number of rounds)
        neg_time_limit: The total time-limit of a single negotiation
        time_limit: The total time-limit of the simulation
        transportation_delay: The transportation delay
        n_retrials: The number of retrials the `Miner` and `GreedyFactoryManager` will try if negotiations fail
        max_insurance_premium: The maximum insurance premium accepted by `GreedyFactoryManager` (-1 to disable)
        use_consumer: If true, the `GreedyFactoryManager` will use an internal consumer for buying its needs
        guaranteed_contracts: If true, the `GreedyFactoryManager` will only sign contracts that it can guaratnee not to
        break.
        consumption_horizon: The number of steps for which `Consumer` publishes `CFP` s
        consumption: The consumption schedule will be sampled from a uniform distribution with these limits inclusive
        log_file_name: File name to store the logs
        negotiator_type: The negotiation factory used to create all negotiators
        max_storage: maximum storage capacity for all factory managers If None then it is unlimited


    Returns:
        SCMLWorld ready to run

    Remarks:

        - Every production level n has one process only that takes n steps to complete


    """
    competitors = list(competitors)
    params = list(params) if params is not None else [dict() for _ in range(len(competitors))]
    if n_factories_per_level == n_default_per_level and len(competitors) > 0:
        raise ValueError(f'All factories in all levels are occupied by the default factory manager. Either decrease'
                         f' n_default_per_level ({n_default_per_level}) or increase n_factories_per_level '
                         f' ({n_factories_per_level})')
    if isinstance(n_intermediate, Iterable):
        n_intermediate = list(n_intermediate)
    else:
        n_intermediate = [n_intermediate, n_intermediate]
    max_insurance_premium = None if max_insurance_premium < 0 else max_insurance_premium
    n_competitors = len(competitors)
    n_intermediate_levels_min = int(math.ceil(n_competitors / (n_factories_per_level - n_default_per_level))) - 1
    if n_intermediate_levels_min > n_intermediate[1]:
        raise ValueError(f'Need {n_intermediate_levels_min} intermediate levels to run {n_competitors} competitors')
    n_intermediate[0] = max(n_intermediate_levels_min, n_intermediate[0])
    competitors = [get_class(c) if isinstance(c, str) else c for c in competitors]
    if len(competitors) < 1:
        competitors.append(GreedyFactoryManager)
        params.append(dict())
    world = SCMLWorld.chain_world(log_file_name=log_file_name, n_steps=n_steps
                                  , agent_names_reveal_type=agent_names_reveal_type
                                  , negotiation_speed=negotiation_speed
                                  , n_intermediate_levels=randint(*n_intermediate)
                                  , n_miners=n_miners
                                  , n_consumers=n_consumers
                                  , n_factories_per_level=n_factories_per_level
                                  , consumption=consumption
                                  , consumer_kwargs={'negotiator_type': negotiator_type
            , 'consumption_horizon': consumption_horizon}
                                  , miner_kwargs={'negotiator_type': negotiator_type, 'n_retrials': n_retrials}
                                  , default_manager_params={'negotiator_type': negotiator_type, 'n_retrials': n_retrials
            , 'sign_only_guaranteed_contracts': guaranteed_contracts
            , 'use_consumer': use_consumer
            , 'max_insurance_premium': max_insurance_premium}
                                  , transportation_delay=transportation_delay
                                  , time_limit=time_limit
                                  , neg_time_limit=neg_time_limit
                                  , neg_n_steps=neg_n_steps
                                  , default_signing_delay=default_signing_delay
                                  , n_lines_per_factory=n_lines_per_factory
                                  , max_storage=max_storage
                                  , manager_types=competitors
                                  , manager_params=params
                                  , n_default_per_level=n_default_per_level
                                  , randomize=randomize
                                  , name=name)

    return world


def balance_calculator(world: SCMLWorld) -> WorldRunResults:
    """A scoring function that scores factory managers' performance by the final balance only ignoring whatever still
    in their inventory.

    Args:
        world: The world which is assumed to be run up to the point at which the scores are to be calculated.

    Returns:
        WorldRunResults giving the names, scores, and types of factory managers.

    """
    result = WorldRunResults(world_name=world.name, log_file_name=world.log_file_name)
    initial_balances = []
    for manager in world.factory_managers:
        if '_default__preassigned__' in manager.id:
            continue
        initial_balances.append(world.a2f[manager.id].initial_balance)
    normalize = all(_ != 0 for _ in initial_balances)
    for manager in world.factory_managers:
        if '_default__preassigned__' in manager.id:
            continue
        factory = world.a2f[manager.id]
        result.names.append(manager.name)
        result.types.append(manager.type_name)
        if normalize:
            result.scores.append((factory.balance - factory.initial_balance) / factory.initial_balance)
        else:
            result.scores.append(factory.balance - factory.initial_balance)
    return result


def anac2019_tournament(competitors: Sequence[Union[str, Type[FactoryManager]]]
                        , randomize=True
                        , agent_names_reveal_type=False
                        , n_agents_per_competitor=1
                        , n_runs_per_config: int = 5
                        , max_n_configs: int = 100
                        , tournament_path: str = './logs/tournaments'
                        , total_timeout: Optional[int] = None
                        , parallelism='parallel'
                        , scheduler_ip: Optional[str] = None
                        , scheduler_port: Optional[str] = None
                        , tournament_progress_callback: Callable[[Optional[WorldRunResults]], None] = None
                        , world_progress_callback: Callable[[Optional[SCMLWorld]], None] = None
                        , name: str = None
                        , verbose: bool = False
                        , configs_only=False
                        , **kwargs
                        ) -> Union[TournamentResults, PathLike]:
    """
    The function used to run ANAC 2019 SCML tournaments.

    Args:

        name: Tournament name
        competitors: A list of class names for the competitors
        randomize: If true, then instead of trying all possible permutations of assignment random shuffles will be used.
        agent_names_reveal_type: If true then the type of an agent should be readable in its name (most likely at its
        beginning).
        n_agents_per_competitor: The number of agents to instantiate in each world from each competitor type.
        max_n_configs: No more than n_runs_max worlds will be run. If `randomize` then it cannot be None and that is exactly
        the number of worlds to run. If not `randomize` then at most this number of worlds will be run if it is not None
        n_runs_per_config: Number of runs per configuration.
        total_timeout: Total timeout for the complete process
        tournament_path: Path at which to store all results. A scores.csv file will keep the scores and logs folder will
        keep detailed logs
        parallelism: Type of parallelism. Can be 'serial' for serial, 'parallel' for parallel and 'distributed' for distributed
        scheduler_port: Port of the dask scheduler if parallelism is dask, dist, or distributed
        scheduler_ip:   IP Address of the dask scheduler if parallelism is dask, dist, or distributed
        world_progress_callback: A function to be called after everystep of every world run (only allowed for serial
        evaluation and should be used with cautious).
        tournament_progress_callback: A function to be called with `WorldRunResults` after each world finished
        processing
        verbose: Verbosity
        configs_only: If true, a config file for each
        kwargs: Arguments to pass to the `world_generator` function

    Returns:
        `TournamentResults` The results of the tournament or a `PathLike` giving the location where configs were saved

    Remarks:
        Default parameters will be used in the league with the exception of `parallelism` which may use distributed
        processing

    """
    return tournament(competitors=competitors, randomize=randomize, agent_names_reveal_type=agent_names_reveal_type
                      , max_n_configs=max_n_configs, n_runs_per_config=n_runs_per_config
                      , tournament_path=tournament_path, total_timeout=total_timeout
                      , parallelism=parallelism, scheduler_ip=scheduler_ip, scheduler_port=scheduler_port
                      , tournament_progress_callback=tournament_progress_callback
                      , world_progress_callback=world_progress_callback, name=name, verbose=verbose
                      , configs_only=configs_only
                      , n_agents_per_competitor=n_agents_per_competitor
                      , world_generator=anac2019_world, score_calculator=balance_calculator, **kwargs)



"""
Implements an agent-world-interface (see `AgentWorldInterface`) for the SCM world.
"""
from typing import Optional, List, Dict, Any

from negmas import Issue
from negmas.apps.scml.common import *
from negmas.java import to_java, from_java
from negmas.situated import AgentWorldInterface, Contract, Action

__all__ = [
    'SCMLAWI',
]


class SCMLAWI(AgentWorldInterface):
    """A single contact point between SCML agents and the world simulation."""

    def register_cfp(self, cfp: CFP) -> None:
        """Registers a CFP"""
        self._world.n_new_cfps += 1
        cfp.money_resolution = self._world.money_resolution
        self.bb_record(section='cfps', key=cfp.id, value=cfp)

    def register_interest(self, products: List[int]) -> None:
        """registers interest in receiving callbacks about CFPs related to these products"""
        self._world.register_interest(agent=self.agent, products=products)

    def unregister_interest(self, products: List[int]) -> None:
        """registers interest in receiving callbacks about CFPs related to these products"""
        self._world.unregister_interest(agent=self.agent, products=products)

    def remove_cfp(self, cfp: CFP) -> bool:
        """Removes a CFP"""
        if self.agent.id != cfp.publisher:
            return False
        return self.bb_remove(section='cfps', key=str(hash(cfp)))

    def evaluate_insurance(self, contract: Contract, t: int = None) -> Optional[float]:
        """Can be called to evaluate the premium for insuring the given contract against breachs committed by others

        Args:

            contract: hypothetical contract
            t: time at which the policy is to be bought. If None, it means current step
        """
        return self._world.evaluate_insurance(contract=contract, agent=self.agent, t=t)

    def buy_insurance(self, contract: Contract) -> bool:
        """Buys insurance for the contract by the premium calculated by the insurance company.

        Remarks:
            The agent can call `evaluate_insurance` to find the premium that will be used.
        """
        return self._world.buy_insurance(contract=contract, agent=self.agent)

    def _create_annotation(self, cfp: 'CFP'):
        """Creates full annotation based on a cfp that the agent is receiving"""
        partners = [self.agent.id, cfp.publisher]
        annotation = {'cfp': cfp, 'partners': partners}
        if cfp.is_buy:
            annotation['seller'] = self.agent.id
            annotation['buyer'] = cfp.publisher
        else:
            annotation['buyer'] = self.agent.id
            annotation['seller'] = cfp.publisher
        return annotation

    def request_negotiation(self, cfp: CFP, req_id: str, roles: List[str] = None, mechanism_name: str = None
                            , mechanism_params: Dict[str, Any] = None) -> bool:
        """
        Requests a negotiation with the publisher of a given CFP

        Args:

            cfp: The CFP to negotiate about
            req_id: A string that is passed back to the caller in all callbacks related to this negotiation
            roles: The roles of the CFP publisher and the agent (in that order). By default no roles are passed (None)
            mechanism_name: The mechanism type to use. If not given the default mechanism from the world will be used
            mechanism_params: Parameters of the mechanism

        Returns:
            Success of failure of the negotiation

        Remarks:

            - The `SCMLAgent` class implements another request_negotiation method that does not receive a `req_id`. This
              helper method is recommended as it generates the required req_id and passes it keeping track of requested
              negotiations (and later of running negotiations). Call this method direclty *only* if you do not
              intend to use the `requested_negotiations` and `running_negotiations` properties of the `SCMLAgent` class

        """
        default_annotation = self._create_annotation(cfp)
        return super().request_negotiation_about(issues=cfp.issues, req_id=req_id,
                                                 partners=default_annotation['partners']
                                                 , roles=roles, annotation=default_annotation,
                                                 mechanism_name=mechanism_name
                                                 , mechanism_params=mechanism_params)

    def request_negotiation_about(self
                                  , issues: List[Issue]
                                  , partners: List[str]
                                  , req_id: str
                                  , roles: List[str] = None
                                  , annotation: Optional[Dict[str, Any]] = None
                                  , mechanism_name: str = None
                                  , mechanism_params: Dict[str, Any] = None
                                  ):
        """
        Overrides the method of the same name in the base class to disable it in SCM Worlds.

        **Do not call this method**

        """
        raise RuntimeError('request_negotiation_about should never be called directly in the SCM world'
                           ', call request_negotiation instead.')

    def is_bankrupt(self, agent_id: str) -> bool:
        """
        Checks whether the given agent is bankrupt

        Args:
            agent_id: Agent ID

        Returns:
            The bankruptcy state of the agent

        """
        return bool(self.bb_read('bankruptcy', key=agent_id))

    def reports_for(self, agent_id: str) -> List[FinancialReport]:
        """
        Gets all financial reports of an agent (in the order of their publication)

        Args:
            agent_id: Agent ID

        Returns:

        """
        reports = self.bb_read('reports_agent', key=agent_id)
        if reports is None:
            return []
        return reports

    def reports_at(self, step: int = None) -> Dict[str, FinancialReport]:
        """
        Gets all financial reports of all agents at a given step

        Args:

            step: Step at which the reports are required. If None, the last set of reports is returned

        Returns:

            A dictionary with agent IDs in keys and their financial reports at the given time as values
        """
        if step is None:
            reports = self.bb_query(section='reports_time', query=None)
            reports = self.bb_read('reports_time', key=str(max([int(_) for _ in reports.keys()])))
        else:
            reports = self.bb_read('reports_time', key=str(step))
        if reports is None:
            return {}
        return reports

    def receive_financial_reports(self, receive: bool = True, agents: Optional[List[str]] = None) -> None:
        """
        Registers/unregisters interest in receiving financial reports

        Args:
            receive: True to receive and False to stop receiving
            agents: If given reception is enabled/disabled only for the given set of agents.

        Remarks:

            - by default financial reports are not sent to any agents. To opt-in to receive financial reports, call this
              method.

        """
        self._world.receive_financial_reports(self.agent, receive, agents)

    @property
    def state(self) -> Factory:
        """Returns the private state of the agent in that world.

        In the SCML world, that is a reference to its factory"""
        return self._world.get_private_state(self.agent)

    @property
    def products(self) -> List[Product]:
        """Products in the world"""
        return self._world.products

    @property
    def processes(self) -> List[Process]:
        """Processes in the world"""
        return self._world.processes

    # sugar functions (implementing actions that can all be done through execute

    def schedule_production(self, profile: int, step: int, contract: Optional[Contract] = None,
                            override: bool = True) -> None:
        """
        Schedules production on the agent's factory

        Args:
            profile: Index of the profile in the agent's `compiled_profiles` list
            step: The step to start production according to the given profile
            contract: The contract for which the production is scheduled (optional)
            override: Whether to override existing production jobs schedules at the same time.

        """
        self.execute(action=Action(type='run', params={'profile': profile, 'time': step
                                                       , 'contract': contract, 'override': override}))

    def stop_production(self, line: int, step: int, contract: Optional[Contract], override: bool = True):
        """
        Stops/cancels production scheduled at the given line at the given time.

        Args:
            line: One of the factory lines (index)
            step: Step to stop/cancel production at
            contract: The contract for which the job is scheduled (optional)
            override: Whether to override existing production jobs schedules at the same time.
        """
        self.execute(action=Action(type='stop', params={'line': line, 'time': step}))

    cancel_production = stop_production
    """
    Stops/cancels production scheduled at the given line at the given time.

    Args:
        line: One of the factory lines (index) 
        step: Step to stop/cancel production at
    """

    def schedule_job(self, job: Job, contract: Optional[Contract]):
        """
        Schedules production using a `Job` object. This can be used to schedule any kind of job

        Args:
            job: The job description
            contract: The contract for which the job is scheduled (optional)

        Remarks:

            - Notice that actions that require the profile member of Job (run) never use the line member and vice versa.
        """
        self.execute(action=Action(type=job.action, params={'profile': job.profile, 'time': job.time
                                                            , 'line': job.line
                                                            , 'contract': contract, 'override': job.override}))

    def hide_inventory(self, product: int, quantity: int) -> None:
        """
        Hides the given quantity of the given product so that it is not accessible by the simulator and does not appear
        in reports etc.

        Args:
            product: product index
            quantity: the amount of the product to hide

        Remarks:

            - if the current quantity in storage of the product is less than the amount to be hidden, whatever quantity
              exists is hidden
            - hiding is always immediate
        """
        self.execute(action=Action(type='hide_product', params={'product': product, 'quantity': quantity}))

    def hide_funds(self, amount: float) -> None:
        """
        Hides the given amount of money so that it is not accessible by the simulator and does not appear
        in reports etc.

        Args:
            amount: The amount of money to hide

        Remarks:

            - if the current cash in the agent's wallet is less than the amount to be hidden, all the cash is hidden.
            - hiding is always immediate
        """
        self.execute(action=Action(type='hide_funds', params={'amount': amount}))

    def unhide_inventory(self, product: int, quantity: int) -> None:
        """
        Un-hides the given quantity of the given product so that it is not accessible by the simulator and does not appear
        in reports etc.

        Args:
            product: product index
            quantity: the amount of the product to hide

        Remarks:

            - if the current quantity in storage of the product is less than the amount to be hidden, whatever quantity
              exists is hidden
            - hiding is always immediate
        """
        self.execute(action=Action(type='unhide_product', params={'product': product, 'quantity': quantity}))

    def unhide_funds(self, amount: float) -> None:
        """
        Un-hides the given amount of money so that it is not accessible by the simulator and does not appear
        in reports etc.

        Args:
            amount: The amount of money to unhide

        Remarks:

            - if the current cash in the agent's wallet is less than the amount to be hidden, all the cash is hidden.
            - hiding is always immediate
        """
        self.execute(action=Action(type='unhide_funds', params={'amount': amount}))


class _ShadowSCMLAWI:
    """An SCMLAWI As seen by JNegMAS.

    This is an object that is not visible to python code. It is not directly called from python ever. It is only called
    from a corresponding Java object to represent an internal python object. Because of he way py4j works, we cannot
    just use dunders to implement this kind of object in general. We will have to implement each such class
    independently.

    This kind of classes will always have an internal Java class implementing a Java interface in Jnegmas that starts
    with Py.

    """

    def __init__(self, awi: SCMLAWI):
        self.awi = awi

    def getProducts(self):
        return to_java(self.awi.products)

    def getProcesses(self):
        return to_java(self.awi.processes)

    def getState(self):
        return to_java(self.awi.state)

    def relativeTime(self):
        return self.awi.relative_time

    def getCurrentStep(self):
        return self.awi.current_step

    def getNSteps(self):
        return self.awi.n_steps

    def getDefaultSigningDelay(self):
        return self.awi.default_signing_delay

    def requestNegotiation(self, cfp, req_id: str, roles=None, mechanism_name=None, mechanism_params=None):
        return self.awi.request_negotiation(from_java(cfp), req_id, roles, mechanism_name, mechanism_params)

    def registerCFP(self, cfp: Dict[str, Any]) -> None:
        """Registers a CFP"""
        self.awi.register_cfp(from_java(cfp))

    def removeCFP(self, cfp: Dict[str, Any]) -> bool:
        """Removes a CFP"""
        return self.awi.remove_cfp(from_java(cfp))

    def registerInterest(self, products: List[int]) -> None:
        """registers interest in receiving callbacks about CFPs related to these products"""
        self.awi.register_interest(from_java(products))

    def unregisterInterest(self, products: List[int]) -> None:
        """registers interest in receiving callbacks about CFPs related to these products"""
        self.awi.unregister_interest(from_java(products))

    def evaluateInsurance(self, contract: Dict[str, Any], t: int = None) -> Optional[float]:
        """Can be called to evaluate the premium for insuring the given contract against breaches committed by others

        Args:

            contract: hypothetical contract
            t: time at which the policy is to be bought. If None, it means current step
        """
        result = self.awi.evaluate_insurance(from_java(contract), t)
        if result < 0:
            return None
        return result

    def buyInsurance(self, contract: Dict[str, Any]) -> bool:
        """Buys insurance for the contract by the premium calculated by the insurance company.

        Remarks:
            The agent can call `evaluate_insurance` to find the premium that will be used.
        """
        return self.awi.buy_insurance(from_java(contract))

    def loginfo(self, msg: str):
        return self.awi.loginfo(msg)

    def logwarning(self, msg: str):
        return self.awi.logwarning(msg)

    def logdebug(self, msg: str):
        return self.awi.logdebug(msg)

    def logerror(self, msg: str):
        return self.awi.logerror(msg)

    class Java:
        implements = ['jnegmas.apps.scml.awi.SCMLAWI']




"""
Genius Negotiator
An agent used to connect to GENIUS agents (ver 8.0.4) and allow them to join negotiation mechanisms

"""
import pathlib

import os
import random
import socket
import subprocess
import time
import typing
from typing import Optional, List, Tuple, Sequence

import pkg_resources
from py4j.java_gateway import (
    JavaGateway, CallbackServerParameters, GatewayParameters)
from py4j.protocol import Py4JNetworkError

from negmas import SAONegotiator, make_discounted_ufun, get_domain_issues, NEGMAS_CONFIG, CONFIG_KEY_GENIUS_BRIDGE_JAR
from negmas import ResponseType, load_genius_domain
from negmas.common import *
from negmas.utilities import UtilityFunction
import json

DEFAULT_JAVA_PORT = 25337
DEFAULT_PYTHON_PORT = 25338

if typing.TYPE_CHECKING:
    from negmas import Outcome

__all__ = [
    'GeniusNegotiator',  # Most abstract kind of agent
    'init_genius_bridge',
    'genius_bridge_is_running'
]

INTERNAL_SEP, ENTRY_SEP, FIELD_SEP = "<<s=s>>", "<<y,y>>", "<<sy>>"

common_gateway: Optional[JavaGateway] = None
common_port: int = 0
java_process = None
python_port: int = 0


all_agent_based_agents = [
    'agents.TestingAgent',
    'agents.SimpleANAC2013Agent',
    'agents.UIAgentExtended',
    'agents.SimilarityAgent',
    'agents.OptimalBidderU',
    'agents.TestAgent',
    'agents.SimpleTFTAgent',
    'agents.BayesianAgentForAuction',
    'agents.ABMPAgent',
    'agents.RandomIncreasingUtilAgent',
    'agents.OptimalBidder',
    'agents.SimpleAgentSavingBidHistory',
    'agents.SimpleAgent',
    'agents.UIAgent',
    'agents.DecUtilAgent',
    'agents.FuzzyAgent',
    'agents.TAgent',
    'agents.anac.y2013.AgentKF.AgentKF',
    'agents.anac.y2013.MetaAgent.portfolio.BRAMAgent2.BRAMAgent2',
    'agents.anac.y2013.MetaAgent.portfolio.IAMhaggler2012.agents2011.SouthamptonAgent',
    'agents.anac.y2013.MetaAgent.portfolio.OMACagent.OMACagent',
    'agents.anac.y2013.MetaAgent.portfolio.AgentLG.AgentLG',
    'agents.anac.y2013.MetaAgent.portfolio.AgentMR.AgentMR',
    'agents.anac.y2013.MetaAgent.portfolio.CUHKAgent.CUHKAgent',
    'agents.anac.y2013.MetaAgent.portfolio.thenegotiatorreloaded.BOAagent',
    'agents.anac.y2013.MetaAgent.MetaAgent2013',
    'agents.anac.y2013.GAgent.AgentI',
    'agents.anac.y2013.SlavaAgent.SlavaAgent',
    'agents.anac.y2013.TMFAgent.TMFAgent',
    'agents.anac.y2014.DoNA.DoNA',
    'agents.anac.y2014.DoNA.ClearDefaultStrategy',
    'agents.anac.y2014.AgentQuest.AgentQuest',
    'agents.anac.y2014.BraveCat.necessaryClasses.BOAagent',
    'agents.anac.y2014.Gangster.Gangster',
    'agents.anac.y2014.Aster.Aster',
    'agents.anac.y2014.AgentTRP.AgentTRP',
    'agents.anac.y2014.E2Agent.AnacSampleAgent',
    'agents.anac.y2014.Sobut.Sobut',
    'agents.anac.y2014.AgentYK.AgentYK',
    'agents.anac.y2014.KGAgent.KGAgent',
    'agents.anac.y2014.SimpaticoAgent.Simpatico',
    'agents.anac.y2014.Atlas.Atlas',
    'agents.anac.y2014.AgentTD.AgentTD',
    'agents.anac.y2014.ArisawaYaki.ArisawaYaki',
    'agents.anac.y2014.Flinch.Flinch',
    'agents.anac.y2014.AgentWhale.WhaleAgent',
    'agents.anac.y2014.AgentM.AgentM',
    'agents.anac.y2012.BRAMAgent2.BRAMAgent2',
    'agents.anac.y2012.IAMhaggler2012.agents2011.SouthamptonAgent',
    'agents.anac.y2012.OMACagent.OMACagent',
    'agents.anac.y2012.AgentLG.AgentLG',
    'agents.anac.y2012.MetaAgent.agents.GYRL.GYRL',
    'agents.anac.y2012.MetaAgent.agents.ShAgent.ShAgent',
    'agents.anac.y2012.MetaAgent.agents.Chameleon.Chameleon',
    'agents.anac.y2012.MetaAgent.agents.SimpleAgentNew.SimpleAgentNew',
    'agents.anac.y2012.MetaAgent.agents.LYY.LYYAgent',
    'agents.anac.y2012.MetaAgent.agents.WinnerAgent.WinnerAgent2',
    'agents.anac.y2012.MetaAgent.agents.DNAgent.DNAgent',
    'agents.anac.y2012.MetaAgent.agents.MrFriendly.MrFriendly',
    'agents.anac.y2012.MetaAgent.MetaAgent',
    'agents.anac.y2012.AgentMR.AgentMR',
    'agents.anac.y2012.CUHKAgent.CUHKAgent',
    'agents.anac.y2017.geneking.GeneKing',
    'agents.anac.y2010.AgentFSEGA.AgentFSEGA',
    'agents.anac.y2010.Yushu.Yushu',
    'agents.anac.y2010.Nozomi.Nozomi',
    'agents.anac.y2010.AgentSmith.AgentSmith',
    'agents.anac.y2010.Southampton.SouthamptonAgentNoExtras',
    'agents.anac.y2010.Southampton.SouthamptonAgentExtrasInterface',
    'agents.anac.y2010.Southampton.SouthamptonAgent',
    'agents.anac.y2010.AgentK.Agent_K',
    'agents.anac.y2011.TheNegotiator.TheNegotiator',
    'agents.anac.y2011.ValueModelAgent.ValueModelAgent',
    'agents.anac.y2011.HardHeaded.KLH',
    'agents.anac.y2011.Gahboninho.Gahboninho',
    'agents.anac.y2011.AgentK2.Agent_K2',
    'agents.anac.y2011.BramAgent.BRAMAgent',
    'agents.anac.y2011.IAMhaggler2011.IAMhaggler2011',
    'agents.anac.y2011.Nice_Tit_for_Tat.BilateralAgent',
    'agents.BayesianAgent',
    'agents.ABMPAgent2',
    'agents.QOAgent',
    'agents.SimpleAgt2',
]
all_party_based_agents = [
    'agents.ai2014.group12.Group12',
    'agents.ai2014.group7.Group7',
    'agents.ai2014.group9.Group9',
    'agents.ai2014.group8.Group8',
    'agents.ai2014.group1.Group1',
    'agents.ai2014.group6.Group6',
    'agents.ai2014.group11.Group11',
    'agents.ai2014.group10.Group10',
    'agents.ai2014.group3.Group3',
    'agents.ai2014.group4.Group4',
    'agents.ai2014.group5.Group5',
    'agents.ai2014.group2.Group2',
    'agents.anac.y2015.Mercury.Mercury',
    'agents.anac.y2015.RandomDance.RandomDance',
    'agents.anac.y2015.cuhkagent2015.CUHKAgent2015',
    'agents.anac.y2015.SENGOKU.SENGOKU',
    'agents.anac.y2015.DrageKnight.DrageKnight',
    'agents.anac.y2015.meanBot.MeanBot',
    'agents.anac.y2015.Phoenix.PhoenixParty',
    'agents.anac.y2015.pokerface.PokerFace',
    'agents.anac.y2015.agenth.AgentH',
    'agents.anac.y2015.ParsAgent.ParsAgent',
    'agents.anac.y2015.AgentHP.AgentHP',
    'agents.anac.y2015.JonnyBlack.JonnyBlack',
    'agents.anac.y2015.xianfa.XianFaAgent',
    'agents.anac.y2015.AresParty.AresParty',
    'agents.anac.y2015.fairy.kawaii',
    'agents.anac.y2015.Atlas3.Atlas3',
    'agents.anac.y2015.TUDMixedStrategyAgent.TUDMixedStrategyAgent',
    'agents.anac.y2015.AgentW.AgentW',
    'agents.anac.y2015.AgentNeo.Groupn',
    'agents.anac.y2015.agentBuyogV2.AgentBuyogMain',
    'agents.anac.y2015.AgentX.AgentX',
    'agents.anac.y2015.pnegotiator.BayesLearner',
    'agents.anac.y2015.pnegotiator.PNegotiator',
    'agents.anac.y2015.group2.Group2',
    'agents.anac.y2017.agentkn.AgentKN',
    'agents.anac.y2017.tucagent.TucAgent',
    'agents.anac.y2017.mosateam.Mosa',
    'agents.anac.y2017.parscat2.ParsCat2',
    'agents.anac.y2017.geneking.GeneKing',
    'agents.anac.y2017.parsagent3.ShahAgent',
    'agents.anac.y2017.mamenchis.Mamenchis',
    'agents.anac.y2017.madagent.MadAgent',
    'agents.anac.y2017.agentf.AgentF',
    'agents.anac.y2017.farma.Farma17',
    'agents.anac.y2017.caduceusdc16.CaduceusDC16',
    'agents.anac.y2017.limitator.Imitator',
    'agents.anac.y2017.simpleagent.SimpleAgent',
    'agents.anac.y2017.rubick.Rubick',
    'agents.anac.y2017.ponpokoagent.PonPokoAgent',
    'agents.anac.y2017.group3.Group3',
    'agents.anac.y2017.gin.Gin',
    'agents.anac.y2017.tangxun.taxibox',
    'agents.anac.y2016.terra.Terra',
    'agents.anac.y2016.maxoops.MaxOops',
    'agents.anac.y2016.grandma.GrandmaAgent',
    'agents.anac.y2016.clockworkagent.ClockworkAgent',
    'agents.anac.y2016.parscat.ParsCat',
    'agents.anac.y2016.caduceus.Caduceus',
    'agents.anac.y2016.caduceus.agents.RandomDance.RandomDance',
    'agents.anac.y2016.caduceus.agents.kawaii.kawaii',
    'agents.anac.y2016.caduceus.agents.Caduceus.Caduceus',
    'agents.anac.y2016.caduceus.agents.ParsAgent.ParsAgent',
    'agents.anac.y2016.caduceus.agents.agentBuyong.agentBuyong',
    'agents.anac.y2016.caduceus.agents.Atlas3.Atlas3',
    'agents.anac.y2016.agentlight.AgentLight',
    'agents.anac.y2016.pars2.ParsAgent2',
    'agents.anac.y2016.yxagent.YXAgent',
    'agents.anac.y2016.farma.Farma',
    'agents.anac.y2016.syagent.SYAgent',
    'agents.anac.y2016.ngent.Ngent',
    'agents.anac.y2016.agenthp2.AgentHP2_main',
    'agents.anac.y2016.atlas3.Atlas32016',
    'agents.anac.y2016.agentsmith.AgentSmith2016',
    'agents.anac.y2016.myagent.MyAgent',
]

agent_based_negotiators = []
party_based_negotiators = [
    # 'agents.ai2014.group12.Group12',
    # 'agents.ai2014.group7.Group7',
    # 'agents.ai2014.group9.Group9',
    # 'agents.ai2014.group1.Group1',
    # 'agents.ai2014.group6.Group6',
    # 'agents.ai2014.group11.Group11',
    # 'agents.ai2014.group10.Group10',
    # 'agents.ai2014.group3.Group3',
    # 'agents.ai2014.group4.Group4',
    # 'agents.ai2014.group5.Group5',
    # 'agents.ai2014.group2.Group2',
    'agents.anac.y2015.Mercury.Mercury',
    'agents.anac.y2015.RandomDance.RandomDance',
    'agents.anac.y2015.cuhkagent2015.CUHKAgent2015',
    'agents.anac.y2015.SENGOKU.SENGOKU',
    'agents.anac.y2015.DrageKnight.DrageKnight',
    'agents.anac.y2015.meanBot.MeanBot',
    'agents.anac.y2015.Phoenix.PhoenixParty',
    'agents.anac.y2015.pokerface.PokerFace',
    'agents.anac.y2015.agenth.AgentH',
    'agents.anac.y2015.ParsAgent.ParsAgent',
    # 'agents.anac.y2015.JonnyBlack.JonnyBlack',
    'agents.anac.y2015.AresParty.AresParty',
    'agents.anac.y2015.fairy.kawaii',
    'agents.anac.y2015.Atlas3.Atlas3',
    'agents.anac.y2015.TUDMixedStrategyAgent.TUDMixedStrategyAgent',
    'agents.anac.y2015.AgentNeo.Groupn',
    'agents.anac.y2015.agentBuyogV2.AgentBuyogMain',
    'agents.anac.y2015.AgentX.AgentX',
    # 'agents.anac.y2015.pnegotiator.BayesLearner',
    # 'agents.anac.y2015.pnegotiator.PNegotiator',
    # 'agents.anac.y2015.group2.Group2',
    'agents.anac.y2017.tucagent.TucAgent',
    'agents.anac.y2017.agentf.AgentF',
    'agents.anac.y2017.ponpokoagent.PonPokoAgent',
    'agents.anac.y2017.tangxun.taxibox',
    'agents.anac.y2016.terra.Terra',
    'agents.anac.y2016.grandma.GrandmaAgent',
    'agents.anac.y2016.clockworkagent.ClockworkAgent',
    'agents.anac.y2016.parscat.ParsCat',
    'agents.anac.y2016.caduceus.Caduceus',
    'agents.anac.y2016.caduceus.agents.RandomDance.RandomDance',
    'agents.anac.y2016.caduceus.agents.kawaii.kawaii',
    'agents.anac.y2016.caduceus.agents.Caduceus.Caduceus',
    'agents.anac.y2016.caduceus.agents.ParsAgent.ParsAgent',
    'agents.anac.y2016.caduceus.agents.agentBuyong.agentBuyong',
    'agents.anac.y2016.caduceus.agents.Atlas3.Atlas3',
    'agents.anac.y2016.pars2.ParsAgent2',
    'agents.anac.y2016.yxagent.YXAgent',
    'agents.anac.y2016.farma.Farma',
    'agents.anac.y2016.syagent.SYAgent',
    'agents.anac.y2016.ngent.Ngent',
    'agents.anac.y2016.atlas3.Atlas32016',
    'agents.anac.y2016.agentsmith.AgentSmith2016',
    'agents.anac.y2016.myagent.MyAgent',
]


def init_genius_bridge(path: str = None, port: int = 0, force: bool = False) -> bool:
    """Initializes a genius connection

    Args:
        path: The path to a JAR file that runs negloader
        port: port number to use
        force: Force trial even if an existing bridge is initialized

    Returns:
        True if successful

    """
    global common_gateway
    global common_port
    global java_process
    global python_port

    port = port if port > 0 else DEFAULT_JAVA_PORT
    if genius_bridge_is_running(port):
        return True
    if not force and common_gateway is not None and common_port == port:
        print('Java already initialized')
        return True

    path = NEGMAS_CONFIG.get(CONFIG_KEY_GENIUS_BRIDGE_JAR, None)
    if path is None:
        print('Cannot find the path to genius bridge jar. Download the jar somewhere in your machine and add its path'
              'to ~/negmas/config.json under the key "genius_bridge_jar".\n\nFor example, if you downloaded the jar'
              ' to /path/to/your/jar then edit ~/negmas/config.json to read something like\n\n'
              '{\n\t"genius_bridge_jar": "/path/to/your/jar",\n\t.... rest of the config\n}\n\n'              
              'You can find the jar at http://www.yasserm.com/scml/genius-8.0.4-bridge.jar')
        return False
    path = pathlib.Path(path).expanduser().absolute()
    try:
        subprocess.Popen(  # ['java', '-jar',  path, '--die-on-exit', f'{port}']
            f'java -jar {path} --die-on-exit {port}'
            , shell=True)
    except (OSError, TimeoutError, RuntimeError, ValueError):
        return False
    time.sleep(0.5)
    gateway = JavaGateway(gateway_parameters=GatewayParameters(port=port),
                          callback_server_parameters=CallbackServerParameters(port=0))
    python_port = gateway.get_callback_server().get_listening_port()
    gateway.java_gateway_server.resetCallbackClient(
        gateway.java_gateway_server.getCallbackClient().getAddress(),
        python_port)

    common_gateway = gateway
    common_port = port
    return True


class GeniusNegotiator(SAONegotiator):
    """Encapsulates a Genius Negotiator"""

    def __init__(self, java_class_name: str
                 , port: int = None
                 , domain_file_name: str = None
                 , utility_file_name: str = None
                 , keep_issue_names: bool = True
                 , keep_value_names: bool = True
                 , auto_load_java: bool = False
                 , can_propose=True
                 , genius_bridge_path: str = None
                 , name: str = None):
        super().__init__(name=name)
        self.capabilities['propose'] = can_propose
        self.add_capabilities({'genius': True})
        self.java = None
        self.java_class_name = java_class_name
        self.port = port
        self.connected = self._connect(path=genius_bridge_path, port=self.port, auto_load_java=auto_load_java)
        self.java_uuid = self._create()
        self.uuid = self.java_uuid
        self.name = self.java_uuid
        self.domain_file_name = str(domain_file_name)
        self.utility_file_name = str(utility_file_name)
        self._my_last_offer = None
        self.keep_issue_names = keep_issue_names
        self.utility_function, self.discount = None, None
        if domain_file_name is not None:
            # we keep original issues details so that we can create appropriate answers to Java
            self.issues = get_domain_issues(domain_file_name=domain_file_name, keep_issue_names=True
                                            , keep_value_names=True)
            self.issue_names = [_.name for _ in self.issues]
            self.issue_index = dict(zip(self.issue_names, range(len(self.issue_names))))
        if utility_file_name is not None:
            self.utility_function, self.discount = UtilityFunction.from_genius(utility_file_name
                                                                               , keep_issue_names=keep_issue_names
                                                                               , keep_value_names=keep_value_names)
        self.base_utility = self.utility_function
        pass

    @classmethod
    def negotiators(cls, agent_based=False, party_based=True) -> List[str]:
        """
        Returns a list of all available agents in genius 8.4.0

        Args:
            agent_based: Old agents based on the Java class Negotiator
            party_based: Newer agents based on the Java class AbstractNegotiationParty

        Returns:

        """
        r = []
        if agent_based:
            r += agent_based_negotiators
        if party_based:
            r += party_based_negotiators
        return r

    @classmethod
    def random_negotiator(cls, agent_based=True, party_based=True
                          , port: int = None
                          , domain_file_name: str = None
                          , utility_file_name: str = None
                          , keep_issue_names: bool = True
                          , keep_value_names: bool = True
                          , auto_load_java: bool = False
                          , can_propose = True
                          , name: str = None
                          ) -> 'GeniusNegotiator':
        """
        Returns an agent with a random class name

        Args:
            agent_based: Old agents based on the Java class Negotiator
            party_based: Newer agents based on the Java class AbstractNegotiationParty


        Returns:
            GeniusNegotiator an agent with a random java class
        """
        agent_names = cls.negotiators(agent_based=agent_based, party_based=party_based)
        agent_name = agent_names[random.randint(0, len(agent_names) - 1)]
        return GeniusNegotiator(java_class_name=agent_name, port=port, domain_file_name=domain_file_name
                                , utility_file_name=utility_file_name, keep_issue_names=keep_issue_names
                                , keep_value_names=keep_value_names
                                , auto_load_java=auto_load_java, can_propose=can_propose
                                , name=name)

    @property
    def is_connected(self):
        return self.connected and self.java is not None

    def _create(self):
        """
        Creates the agent
        Returns:

        Examples:
            >>> if genius_bridge_is_running():
            ...     a = GeniusNegotiator(java_class_name="agents.anac.y2015.Atlas3.Atlas3")
            ...     a.java_uuid.startswith("agents.anac.y2015.Atlas3.Atlas3")
            ... else:
            ...     True
            True
            >>> if genius_bridge_is_running():
            ...     len(a.java_uuid)- len(a.java_class_name) == 36 # length of UUID
            ... else:
            ...     True
            True

        """
        return self.java.create_agent(self.java_class_name)

    def _connect(self, path: str, port: int, auto_load_java: bool = False) -> bool:
        """

        Returns:

        Examples:

            - Testing multilateral agent
            >>> if genius_bridge_is_running():
            ...     a = GeniusNegotiator(java_class_name="agents.anac.y2015.Atlas3.Atlas3")
            ...     print(a.java_name)
            ... else:
            ...     print('ANAC2015-6-Atlas')
            ANAC2015-6-Atlas

            - Testing bilateral agent
            >>> if genius_bridge_is_running():
            ...    b = GeniusNegotiator(java_class_name="agents.SimpleAgent")
            ...    print(b.java_name)
            ... else:
            ...    print('Agent SimpleAgent')
            Agent SimpleAgent

        """
        if port is None:
            if auto_load_java:
                if common_gateway is None:
                    init_genius_bridge(path=path)
                gateway = common_gateway
                self.java = gateway.entry_point
                port = DEFAULT_JAVA_PORT
                return True
            else:
                port = DEFAULT_JAVA_PORT
        gateway = JavaGateway(gateway_parameters=GatewayParameters(port=port, auto_close=True))
        if gateway is None:
            self.java = None
            return False
        self.java = gateway.entry_point
        return True

    @property
    def java_name(self):
        return self.java.getName(self.java_uuid)

    def test(self) -> str:
        return self.java.test(self.java_class_name)

    def on_negotiation_start(self, state: MechanismState) -> None:
        """Called when the info starts. Connects to the JVM.
        """
        super().on_negotiation_start(state=state)
        info = self._ami
        if self.discount is not None and self.discount != 1.0:
            self.utility_function = make_discounted_ufun(self.utility_function, info=info
                                                , discount_per_round=self.discount, power_per_round=1.0)
        n_steps = -1 if info.n_steps is None else int(info.n_steps)  # number of steps
        n_seconds = -1 if info.time_limit is None else int(info.time_limit)  # time limit
        if n_steps * n_seconds > 0:
            # n_seconds take precedence
            n_steps = -1
        self.java.on_negotiation_start(
            self.java_uuid  # java_uuid
            , info.n_negotiators  # number of agents
            , n_steps
            , n_seconds
            , n_seconds > 0
            , self.domain_file_name  # domain file
            , self.utility_file_name  # Negotiator file
        )

    def propose(self, state: MechanismState) -> 'Outcome':
        if not self.capabilities['propose']:
            return None
        if self._my_last_offer is None:  # never responded before
            response, outcome = self.parse(self.java.choose_action(self.java_uuid))
            if outcome is None:
                return None
            self._my_last_offer = outcome
            return outcome

        # we offered something before
        tmp = self._my_last_offer
        self._my_last_offer = None
        return tmp

    def respond(self, state: MechanismState, offer: 'Outcome') -> 'ResponseType':
        action = self.java.choose_action(self.java_uuid)
        response, self._my_last_offer = self.parse(action)
        return response

    def parse(self, action: str) -> Tuple[Optional[ResponseType], Optional['Outcome']]:
        """
        Parses an action into and a ResponseType and an Outcome (if one is included)
        Args:
            action:

        Returns:

        """
        response, outcome = None, None
        id, typ_, bid_str = action.split(FIELD_SEP)
        if typ_ in ('Offer',) and (bid_str is not None and len(bid_str) > 0):
            try:
                if self.keep_issue_names:
                    outcome = {_[0]: _[1] for _ in [_.split(INTERNAL_SEP) for _ in bid_str.split(ENTRY_SEP)]}
                else:
                    outcome = tuple(_.split(INTERNAL_SEP)[1] for _ in bid_str.split(ENTRY_SEP))
            except:
                print(f'Failed for bid string: {bid_str} of action {action}')
        if typ_ == 'Offer':
            response = ResponseType.REJECT_OFFER
        elif typ_ == 'Accept':
            response = ResponseType.ACCEPT_OFFER
        elif typ_ == 'EndNegotiation':
            response = ResponseType.END_NEGOTIATION
        else:
            raise ValueError(f'Unknown response: {typ_} in action {action}')
        return response, outcome

    def _outcome2str(self, outcome):
        output = ''
        if not isinstance(outcome, dict):
            outcome_dict = dict(zip(self.issue_names, outcome))
        else:
            outcome_dict = outcome
        for i, v in outcome_dict.items():
            output += f'{i}{INTERNAL_SEP}{v}{ENTRY_SEP}'
        output = output[:-len(ENTRY_SEP)]
        return output

    def __str__(self):
        name = super().__str__().split('/')
        return '/'.join(name[:-1]) + f'/{self.java_class_name}/' + name[-1]

    def on_partner_proposal(self, state: MechanismState, agent_id: str, offer: 'Outcome'):
        if agent_id is self.id:
            return
        agent_info = [_ for _ in self._ami.participants if _.id != self.id and _.id == agent_id]
        if len(agent_info) == 0:
            return
        agent_info = agent_info[0]
        #if agent_info.type == 'genius_negotiator':
        #    return
        self.java.receive_message(self.java_uuid
                                  , agent_id, 'Offer'
                                  , self._outcome2str(offer))

    def on_partner_response(self, state: MechanismState, agent_id: str, outcome: 'Outcome', response: 'ResponseType'):
        if agent_id is self.id:
            return
        agent_info = [_ for _ in self._ami.participants if _.id != self.id and _.id == agent_id]
        if len(agent_info) == 0:
            return
        if outcome is None:
            return
        agent_info = agent_info[0]
        #if agent_info.type == 'genius_negotiator':
        #    return
        bid = self._outcome2str(outcome)
        if response == ResponseType.END_NEGOTIATION:
            resp = 'EndNegotiation'
        elif response == ResponseType.ACCEPT_OFFER:
            resp = 'Accept'
        elif response == ResponseType.REJECT_OFFER:
            resp = 'Reject'
        else:
            return
        self.java.receive_message(self.java_uuid
                                  , agent_id
                                  , resp, bid)


def genius_bridge_is_running(port: int = None) -> bool:
    """
    Checks whether a Genius Bridge is running. A genius bridge allows you to use `GeniusNegotiator` objects.

    Remarks:

        You can start a Genius Bridge in at least two ways:

        - execute the python function `init_genius_bridge()` in this module
        - run "negmas genius" on the terminal

    """
    if port is None:
        port = DEFAULT_JAVA_PORT
    s = socket.socket()
    try:
        s.connect(('127.0.0.1', port))
        return True
    except ConnectionRefusedError:
        return False
    except IndexError:
        return False
    except Py4JNetworkError:
        return False
    finally:
        s.close()




from random import gauss
import numpy as np

__all__ = [
    'pos_gauss', 'safe_max', 'zero_runs'
]


def pos_gauss(mu, sigma):
    """Returns a sample from a rectified gaussian"""
    x = gauss(mu, sigma)
    return abs(x)


def safe_max(a, b):
    if a is None:
        return b
    if b is None:
        return a
    return max(a, b)


def zero_runs(a: np.array) -> np.array:
    """
    Finds all runs of zero in an array

    Args:
        a: Input array (assumed to be 1D)

    Returns:
        np.array: A 2D array giving beginning and end (exclusive) of zero stretches in the input array.
    """
    if len(a) == 0:
        return []
    if np.all(np.equal(a, 0).view(np.int8)):
        return np.array([[0, len(a)]])
    iszero = np.concatenate(([0], np.equal(a, 0).view(np.int8), [0]))
    absdiff = np.abs(np.diff(iszero))
    # Runs start and end where absdiff is 1.
    ranges = np.where(absdiff == 1)[0].reshape(-1, 2)
    if len(ranges) == 0 and a[0] == 0:
        return np.array([[0, len(a)]])
    return ranges



"""
Implements Java interoperability allowing parts of negmas to work smoothly with their Java counterparts in jnegmas

"""
import os
import socket
import subprocess
import time
from contextlib import contextmanager
from typing import Optional, Dict, Any, Iterable, Union
import numpy as np

import pkg_resources
from py4j.clientserver import ClientServer, JavaParameters, PythonParameters
from py4j.java_collections import ListConverter, JavaMap, MapConverter, JavaList, JavaSet
from py4j.java_gateway import JavaGateway, GatewayParameters, CallbackServerParameters, JavaObject
# @todo use launch_gateway to start the java side. Will need to know the the jar location so jnegmas shoud save that
#  somewhere
from py4j.protocol import Py4JNetworkError

from negmas.helpers import get_class, instantiate, camel_case, snake_case

__all__ = [
    'JavaCallerMixin',
    'JNegmasGateway',
    'to_java',
    'from_java',
    'jnegmas_bridge_is_running',
    'init_jnegmas_bridge',
    'jnegmas_connection',
    'from_java',
    'java_link',
    'to_dict_for_java',
]

DEFAULT_JNEGMAS_PATH = 'external/jnegmas-1.0-SNAPSHOT-all.jar'
PYTHON_CLASS_IDENTIFIER = '__python_class__'


@contextmanager
def jnegmas_connection(init: bool = False, path: Optional[str] = None, java_port=0
                       , python_port=0, client_server=True, shutdown=True):
    """A connection to jnegmas that closes automatically"""
    if init:
        JNegmasGateway.start_java_side(path=path, java_port=java_port)
    JNegmasGateway.connect(java_port=java_port, python_port=python_port, client_server=client_server)
    yield JNegmasGateway.gateway
    if shutdown:
        JNegmasGateway.shutdown()


def init_jnegmas_bridge(path: Optional[str] = None, port: int = 0):
    JNegmasGateway.start_java_side(path=path, java_port=port)


def jnegmas_bridge_is_running(port: int = None) -> bool:
    """
    Checks whether a JNegMAS Bridge is running. This bridge is needed to use any objects in the jnegmas package

    Remarks:

        You can start a JNegMAS Bridge in at least two ways:

        - execute the python function `init_jnegmas_bridge()` in this module
        - run "negmas jnegmas" on the terminal

    """
    return JNegmasGateway.is_running(port)


class PyEntryPoint:
    """Used as an entry point from the java side allowing it to create objects in python and take initiative in calling
    python"""

    def create(self, class_name: str, params: Dict[str, Any]) -> Any:
        """
        Creates a python object and returns it.

        Args:
            class_name: python full class name
            params: params to pass by value to the constructor of the object

        Remarks:
            - Notice that the returned object will only be usable in Java if it is implementing a Java interface which
              means that it *must* have an internal Java class with an `implements` list of interfaces that it
              implements.

        """
        return from_java(params, fallback_class_name=class_name)

    def create_shadow(self, class_name: str, params: Dict[str, Any]) -> Any:
        lst = class_name.split('.')
        for ending in ('Negotiator', 'FactoryManager', 'UtilityFunction'):
            if lst[-1].endswith(ending):
                lst[-1] = '_Shadow' + (ending if ending != 'Negotiator' else 'SAONegotiator')
                break
        else:
            lst[-1] = '_Shadow' + lst[-1]
        shadow_class_name = '.'.join(lst)
        cls = get_class(shadow_class_name)
        obj = from_java(params, fallback_class_name=class_name)
        shadow = cls(obj)
        return shadow

    class Java:
        implements = ['jnegmas.EntryPoint']


class JNegmasGateway:
    DEFAULT_PYTHON_PORT = 25334
    DEFAULT_JAVA_PORT = 25333

    gateway: Optional[Union[JavaGateway, ClientServer]] = None

    @classmethod
    def start_java_side(cls, path: str = None, java_port: int = 0) -> None:
        """Initializes a connection to jnegmas

        Examples:

            # >>> start_java_side(port=35337)
            # >>> a = JNegmasGateway.do_nothing_manager()
            # >>> a.java_uuid.startswith('jnegmas')
            # True
            # >>> len(a.java_uuid)- len(a.java_class_name) == 36 # length of UUID
            # True

        """
        if cls.gateway is not None:
            return
        if path is None:
            path = pkg_resources.resource_filename('negmas', resource_name=DEFAULT_JNEGMAS_PATH)
        java_port = java_port if java_port > 0 else cls.DEFAULT_JAVA_PORT
        if jnegmas_bridge_is_running(port=java_port):
            return
        path = os.path.abspath(os.path.expanduser(path))
        try:
            subprocess.Popen(  # ['java', '-jar',  path, '--die-on-exit', f'{port}']
                f'java -jar {path} --doe --client-server --port {java_port}'
                , shell=True)
        except FileNotFoundError:
            print(os.getcwd(), flush=True)
            raise FileNotFoundError([os.getcwd(), path])
        except:
            pass
        time.sleep(0.5)
        cls.connect(auto_load_java=False)

    @classmethod
    def connect(cls, java_port: int = None, python_port: int = None
                , auto_load_java: bool = False, client_server: bool = True) -> bool:
        """
        connects to jnegmas
        """
        if not java_port:
            java_port = cls.DEFAULT_JAVA_PORT
        if not python_port:
            python_port = cls.DEFAULT_PYTHON_PORT
        if auto_load_java:
            if cls.gateway is None:
                cls.start_java_side(java_port=java_port)
            return True
        if cls.gateway is None:
            eager_load, auto_convert = True, True
            auto_field, auto_gc = True, False
            if client_server:
                cls.gateway = ClientServer(java_parameters=JavaParameters(port=java_port
                                                                          , auto_convert=auto_convert
                                                                          , eager_load=eager_load, auto_close=True
                                                                          , auto_gc=False, auto_field=auto_field
                                                                          , daemonize_memory_management=True),
                                           python_parameters=PythonParameters(port=python_port
                                                                              , propagate_java_exceptions=True
                                                                              , daemonize=True
                                                                              , eager_load=eager_load
                                                                              , auto_gc=auto_gc
                                                                              , daemonize_connections=True
                                                                              ),
                                           python_server_entry_point=PyEntryPoint())
            else:
                pyparams = CallbackServerParameters(port=python_port
                                                    , daemonize_connections=True
                                                    , daemonize=True
                                                    , eager_load=eager_load
                                                    , propagate_java_exceptions=True)
                cls.gateway = JavaGateway(gateway_parameters=GatewayParameters(port=java_port
                                                                               , auto_convert=auto_convert
                                                                               , auto_field=auto_field
                                                                               , eager_load=eager_load
                                                                               , auto_close=True),
                                          callback_server_parameters=pyparams, auto_convert=auto_convert
                                          , start_callback_server=True, eager_load=eager_load
                                          , python_server_entry_point=PyEntryPoint())
            python_port = cls.gateway.get_callback_server().get_listening_port()
            cls.gateway.java_gateway_server.resetCallbackClient(
                cls.gateway.java_gateway_server.getCallbackClient().getAddress(),
                python_port)
        return True

    @classmethod
    def shutdown(cls):
        cls.gateway.shutdown(raise_exception=False)
        cls.gateway.shutdown_callback_server(raise_exception=False)

    @classmethod
    def is_running(cls, port):
        """
        Checks whether a JNegMAS Bridge is running. This bridge is needed to use any objects in the jnegmas package

        Remarks:

            You can start a JNegMAS Bridge in at least two ways:

            - execute the python function `init_jnegmas_bridge()` in this module
            - run "negmas jnegmas" on the terminal

        """
        if port is None:
            port = JNegmasGateway.DEFAULT_JAVA_PORT
        s = socket.socket()
        try:
            s.connect(('127.0.0.1', port))
            return True
        except ConnectionRefusedError:
            return False
        except IndexError:
            return False
        except Py4JNetworkError:
            return False
        except Exception:
            return False
        finally:
            s.close()


def java_identifier(s: str):
    if s != PYTHON_CLASS_IDENTIFIER:
        return camel_case(s)
    return s


def to_dict_for_java(value, deep=True, add_type_field=True):
    """Encodes the given value as nothing not more complex than simple dict of either dicts, lists or builtin numeric
    or string values

    Args:
        value: Any object
        deep: Whether we should go deep in the encoding or do a shallow encoding
        add_type_field: Whether to add a type field. If True, A field named `PYTHON_CLASS_IDENTIFIER` will be added
        giving the type of `value`

    Remarks:

        - All iterables are converted to lists when `deep` is true.
        - If the `value` object has a `to_java` member, it will be called to do the conversion, otherwise its `__dict__`
          or `__slots__` member will be used.

    See Also:
          `from_java`, `PYTHON_CLASS_IDENTIFIER`

    """
    def good_field(k: str):
        return not k.startswith('python_') and not k.startswith('java') and not (k != PYTHON_CLASS_IDENTIFIER and k.startswith('_'))
    if isinstance(value, JavaObject):
        return value
    if isinstance(value, dict):
        if not deep:
            return {java_identifier(k): v for k, v in value.items()}
        return {java_identifier(k): to_dict_for_java(v) for k, v in value.items() if good_field(k)}
    if isinstance(value, Iterable) and not deep:
        return value
    if isinstance(value, Iterable) and not isinstance(value, str):
        return [to_dict_for_java(_) for _ in value]
    if hasattr(value, 'to_java'):
        converted = value.to_java()
        if isinstance(converted, dict):
            if add_type_field:
                converted[PYTHON_CLASS_IDENTIFIER] = value.__class__.__module__ + '.' + value.__class__.__name__
            return {java_identifier(k): v for k, v in converted.items()}
        else:
            return converted
    if hasattr(value, '__dict__'):
        if deep:
            d = {java_identifier(k): to_dict_for_java(v) for k, v in value.__dict__.items()
                 if good_field(k)}
        else:
            d = {java_identifier(k): v for k, v in value.__dict__.items()
                 if good_field(k)}
        if add_type_field:
            d[PYTHON_CLASS_IDENTIFIER] = value.__class__.__module__ + '.' + value.__class__.__name__

        # ugly ugly ugly ugly
        if '_NamedObject__uuid' in value.__dict__:
            d['id'] = value.__dict__['_NamedObject__uuid']
        if '_NamedObject__name' in value.__dict__:
            d['name'] = value.__dict__['_NamedObject__name']

        return d
    if hasattr(value, '__slots__'):
        if deep:
            d = dict(zip((camel_case(k) for k in value.__slots__), (to_dict_for_java(getattr(value, _)) for _ in value.__slots__)))
        else:
            d = dict(zip((camel_case(k) for k in value.__slots__), (getattr(value, _) for _ in value.__slots__)))
        if add_type_field:
            d[PYTHON_CLASS_IDENTIFIER] = value.__class__.__module__ + '.' + value.__class__.__name__
        return d
    # a builtin
    if isinstance(value, np.int64):
        return int(value)
    return value


def java_link(obj, map=None, copyable=False):
    """
    Creates a link in java to the object given without copying it.

    Args:
        obj: The object for which to create a java shadow
        map: construction parameters
        copyable: If true, we will assume that the java object is PyCopyable otherwise PyConstructable. Only checked if
        map is not None

    Returns:
        A java object. Cannot be used dieectly in python but can be used as an argument to a call to of a java object.

    """
    class_name = obj.__class__.__module__ + '.' + obj.__class__.__name__
    lst = class_name.split('.')
    lst[0] = 'j' + lst[0]
    if lst[-1].startswith('_Shadow'):
        lst[-1] = lst[-1][len('_Shadow'):]
    if lst[-1].startswith('Shadow'):
        lst[-1] = lst[-1][len('Shadow'):]
    lst[-1] = 'Python' + lst[-1]
    class_name = '.'.join(lst)
    java_obj = JNegmasGateway.gateway.entry_point.create(class_name, obj)
    if map is not None:
        if copyable:
            java_obj.fromMap(map)
        else:
            java_obj.construct(map)
    return java_obj


def to_java(value, add_type_field=True, python_class_name: str = None):
    """Encodes the given value as nothing not more complex than simple dict of either dicts, lists or builtin numeric
    or string values

    Args:
        value: Any object
        add_type_field: If true, the `PYTHON_CLASS_IDENTIFIER`  will be added with the python class field on it
        python_class_name: It given it overrides the class name written when `add_type_field` is True otherwise, the
        class name will be inferred as the __class__ of `value`.

    Remarks:

        - All iterables are converted to lists when `deep` is true.
        - If the `value` object has a `to_java` member, it will be called to do the conversion, otherwise its `__dict__`
          or `__slots__` member will be used.

    See Also:
          `from_java`, `PYTHON_CLASS_IDENTIFIER`

    """
    if value is None:
        return None
    value = to_dict_for_java(value, deep=True, add_type_field=add_type_field)
    if isinstance(value, JavaObject):
        return value
    if isinstance(value, np.int64):
        return int(value)
    if isinstance(value, dict):
        for k, v in value.items():
            if isinstance(v, np.int64):
                value[k] = int(v)
        if add_type_field and python_class_name is not None:
            value[PYTHON_CLASS_IDENTIFIER] = python_class_name
        return JNegmasGateway.gateway.entry_point.createJavaObjectFromMap(value)

    if isinstance(value, Iterable) and not isinstance(value, str):
        return ListConverter().convert([JNegmasGateway.gateway.entry_point.createJavaObjectFromMap(_)
                                        if isinstance(_, dict) else _ for _ in value]
                                       , JNegmasGateway.gateway._gateway_client)
    raise RuntimeError(f'got {value} of type {type(value)} from to_dict_fo_java')


def from_java(d: Any, deep=True, remove_type_field=True, fallback_class_name: Optional[str] = None):
    """Decodes a dict coming from java recovering all objects in the way

    Args:
        d: The value to be decoded. If it is not a dict, it is returned as it is.
        deep: If true, decode recursively
        remove_type_field: If true the field called `PYTHON_CLASS_IDENTIFIER` will be removed if found.
        fallback_class_name: If given, it is used as the fall-back  type if ``PYTHON_CLASS_IDENTIFIER` is not in the dict.

    Remarks:

        - If the object is not a dict or if it has no `PYTHON_CLASS_IDENTIFIER` field and no `fallback_class_name` is
          given, the input `d` is returned as it is. It will not even be copied.

    See Also:
        `to_java`, `PYTHON_CLASS_IDENTIFIER`



    """
    def good_field(k: str):
        return not k.startswith('python_') and not k.startswith('java') and not (k != PYTHON_CLASS_IDENTIFIER and k.startswith('_'))
    if d is None or isinstance(d, int) or isinstance(d, float) or isinstance(d, str):
        return d
    if isinstance(d, JavaList):
        return [_ for _ in d]
    if isinstance(d, JavaSet):
        return {_ for _ in d}
    if isinstance(d, JavaMap):
        d = {snake_case(k): d[k] for k in d.keys()}
    if isinstance(d, JavaObject):
        d = JNegmasGateway.gateway.entry_point.toMap(d)
        d = {k: d[k] for k in d.keys()}
    if isinstance(d, dict):
        if remove_type_field:
            python_class_name = d.pop(PYTHON_CLASS_IDENTIFIER, fallback_class_name)
        else:
            python_class_name = d.get(PYTHON_CLASS_IDENTIFIER, fallback_class_name)
        if python_class_name is not None:
            lst = python_class_name.split('.')
            if lst[-1].startswith('Python'):
                lst[-1] = lst[-1][len('Python'):]
            python_class_name = '.'.join(lst)
            python_class = get_class(python_class_name)
            # we resolve sub-objects first from the dict if deep is specified before calling from_java on the class
            if deep:
                d = {snake_case(k): from_java(v) for k, v in d.items() if good_field(k)}
            # from_java needs to do a shallow conversion from a dict as deep conversion is taken care of already.
            if hasattr(python_class, 'from_java'):
                return python_class.from_java({snake_case(k): v for k, v in d.items()})
            return python_class(**{snake_case(k): v for k, v in d.items() if good_field(k)})
        return d
    raise(ValueError(str(d)))


class JavaCallerMixin:
    """A mixin to enable calling a java counterpart. This mixin can ONLY be used with a `NamedObject` because it uses
    its id property.

    Other than inheriting this mixin, you should call its `init_java_bridge` in your `__init__` (or whenever your object
    is initialized and you need to create the Java counterpart). You should then implement all your functions as calls
    to `java_object`.

    If for example you have a function `do_this`, you can just define it as follows:

    .. code

        def do_this(self, x, y, z) -> Type:
            return self.java_object.doThis(x, y, z)

    Notice that you cannot use named arguments when calling the function in `java_object` and that the names are
    converted to camelCase instead of snake_case. Moreover, property x will be implemented as a pair getX, setX on the
    Java side.

    If your class needs just to call the corresponding java object but is never called back from it then you are done
    after inheriting from this mixin.

    If your objects need to be called from the java counterpart, then you need to add the following to your class
    definition

    .. code

        class Java:
            implements = ['jnegmas.ClassName']

    This assumes that your class is named `ClassName` and that there is an interface called `ClassName` defined in
    jnegmas that has the *same* public interface as your class (or whatever part of it to be called from Java).

    """

    @classmethod
    def from_dict(cls, java_object, *args, **kwargs):
        """Creates a Python object representing the corresponding Java object"""
        obj = cls(*args, **kwargs)
        obj._java_object = java_object
        obj._connected = True
        obj._java_class_name = java_object.getClass().getSimpleName()
        return obj

    def init_java_bridge(self, java_object, java_class_name: str, auto_load_java: bool = False
                         , python_shadow_object: Any = None):
        """
        initializes a connection to the java bridge creating a member called `java_object` that can be used to access
        the counterpart object in Java

        Args:
            java_object: A java object that already exists of the correct type. If given no new objects will be created
            java_class_name: The type of the Java object to be created
            auto_load_java: When true, a JVM will be automatically created (if one is not available)
            python_shadow_object: A python object to shadow the java object. The object will just call the corresponding
            method on this shadow object whenever it needs.


        Remarks:

            - sets a member called java_object that can be used to access the corresponding Java object crated
            - if `python_shadow_object` is given, it must be an object of a type that has an internal class called
              Java which has a single member called 'implements' which is a list of one string element
              representing the Java interface being implemented (it must be either jnegmas.PyCallable or an extension of
              it).
        """
        self._java_class_name = java_class_name if java_class_name is not None else java_object.getClass().getName()
        self._connected = JNegmasGateway.connect(auto_load_java=auto_load_java)
        self._java_object = java_object if java_object is not None else self._create(python_shadow_object)

    def _create(self, python_shadow: Any = None):
        """
        Creates the internal object.

        Args:

              python_shadow: An object that has an internal Java class with implements = ['jnegmas.Callable']

        """
        if python_shadow is None:
            return JNegmasGateway.gateway.entry_point.create(self._java_class_name)
        return JNegmasGateway.gateway.entry_point.create(self._java_class_name, python_shadow)



from random import seed
#seed(0)
import pytest
from pprint import pprint
import time
from negmas import UtilityDistribution as U
import negmas

start = time.monotonic()
import os
from negmas import HyperRectangleUtilityFunction
from negmas import LimitedOutcomesNegotiator
from negmas import SAOMechanism
print(f'Import took {time.monotonic()-start}')


def test_a_session():
    start = time.monotonic()
    p = SAOMechanism(outcomes=50, n_steps=50)
    for _ in range(4):
        p.add(LimitedOutcomesNegotiator(outcomes=50, p_ending=0.01, name=f'agent {_}')
              , ufun=HyperRectangleUtilityFunction([None], [lambda x: x[0]]))
    p.run()
    # print(f'{len(p.negotiators)} negotiators')
    assert len(p.history) > 0
    # print(f'Took {time.monotonic()-start}')


if __name__ == '__main__':
    pytest.main(args=[__file__])



import sys
from enum import Enum

import pytest

from negmas.helpers import create_loggers, unique_name, is_nonzero_file, ConfigReader
from negmas.helpers import pretty_string


def test_create_loggers_with_default_params(capsys):
    log = create_loggers()

    log.info('Test info')
    _, captured = capsys.readouterr()
    assert captured == ''


def disabled_test_create_loggers_with_file_params(capsys, tmpdir):
    file_name = tmpdir.join("log.txt")

    log = create_loggers(file_name=file_name, module_wide_log_file=True, app_wide_log_file=False)

    log.info('Test info')
    #_, captured = capsys.readouterr()
    #assert captured == ''

    # with open(file_name, 'r') as f:
    #     flog = f.read()
    #     assert flog.endswith('INFO - Test info\n')

    log.warning('Test message')
    #_, captured = capsys.readouterr()
    #assert captured.endswith('WARNING - Test message\n')

    # with open(file_name, 'r') as f:
    #     flog = f.read()
    #     assert flog.endswith('WARNING - Test message\n')


def test_unique_name_defaults():
    a = unique_name('')
    assert len(a) == 8 + 1 + 6 + 8


def test_unique_name_no_time():
    assert len(unique_name('', add_time=False)) == 8


def test_unique_name_with_path(tmpdir):
    a = unique_name(str(tmpdir))
    assert a.startswith(str(tmpdir))


def test_is_nonzero_file(tmpdir):
    f_name = unique_name('')
    f = tmpdir / f_name
    assert is_nonzero_file(str(f)) is False

    with open(f, 'w') as tst_file:
        tst_file.write('')
    assert is_nonzero_file(str(f)) is False

    with open(f, 'w') as tst_file:
        tst_file.write('test')
    assert is_nonzero_file(str(f)) is True


def test_pretty_string(capsys):

    assert pretty_string('Test with no components') == 'Test with no components'
    assert pretty_string(4) == '4'
    assert pretty_string(2.5) == '2.5'
    assert pretty_string([1, 2, 3]) == '[\n  1\n  2\n  3\n]'
    assert pretty_string((1, 2, 3)) == '[\n  1\n  2\n  3\n]'
    assert pretty_string({'a': 1, 'b': 2}) == '{\n  a:1\n  b:2\n}'

    assert pretty_string([1, 2, 3], compact=True) == '[  1  2  3]'
    assert pretty_string((1, 2, 3), compact=True) == '[  1  2  3]'
    assert pretty_string({'a': 1, 'b': 2}, compact=True) == '{  "a":1  "b":2}'


def test_config_reader():
    class A(ConfigReader):
        def __init__(self, a, b):
            self.a = a
            self.b = b

    test, children = A.from_config(config={'a': 10, 'b': [1, 2, 3]}
                                   , ignore_children=False, scope=locals())
    assert isinstance(test, A)
    assert test.a == 10
    assert test.b == [1, 2, 3]


def test_config_reader_with_enum():
    class E(Enum):
        E1 = 1
        E2 = 2

    class A(ConfigReader):
        def __init__(self, a, b, e: E):
            self.a = a
            self.b = b
            self.e = e

    test, children = A.from_config(config={'a': 10, 'b': [1, 2, 3], 'e:E': 1}
                                   , ignore_children=False, scope=locals())
    assert isinstance(test, A)
    assert test.a == 10
    assert test.b == [1, 2, 3]
    assert test.e == E.E1


def test_config_reader_ignoring_children():
    class A(ConfigReader):
        def __init__(self, a, b):
            self.a = a
            self.b = b

    test = A.from_config(config={'a': 10, 'b': [1, 2, 3]}, scope=locals())
    assert isinstance(test, A)
    assert test.a == 10
    assert test.b == [1, 2, 3]


class B(ConfigReader):
    def __init__(self, bb, bbdefault=5):
        self.bb = bb
        self.bbdefaults = bbdefault


class Other(ConfigReader):
    def __init__(self, o):
        self.o = o


class C(ConfigReader):
    def __init__(self, cc, b: B=None, others=None):
        self.cc = cc
        self.b = B
        self.others = others

    def set_others(self, others):
        self.others = others


class A(ConfigReader):
    def __init__(self, a, b=None, c=None):
        self.a = a
        self.b = b
        self.c = c


def test_config_reader_with_subobjects():
    test = A.from_config(config={
        'a': 10,
        'b': {'bb': 20},
        'c': {'cc': 5,
              'b': {'bb': 20},
              'others': [{'o': 1}, {'o': 2}, None, {'o': 3}],
              }
    }, scope=globals()
    )

    assert isinstance(test, A)
    assert test.a == 10
    assert isinstance(test.b, B)
    assert isinstance(test.c, C)
    assert isinstance(test.c.others, list)
    assert isinstance(test.c.others[0], Other)
    assert len(test.c.others) == 4
    assert test.c.others[2] is None


if __name__ == '__main__':
    pytest.main(args=[__file__])



import math
from abc import ABC, abstractmethod
from typing import Dict, Any, Callable, Collection, List, Optional

import numpy as np
from dataclasses import dataclass, field

from negmas.apps.scml.helpers import zero_runs
from negmas.situated import Contract
from .common import ProductionNeed, Job, Product, Process, ProductManufacturingInfo, SCMLAgreement, \
    ManufacturingProfileCompiled, INVALID_UTILITY
from .simulators import FactorySimulator, transaction
from .common import NO_PRODUCTION

__all__ = [
    'ScheduleInfo', 'Scheduler', 'GreedyScheduler'
]


@dataclass
class ScheduleInfo:
    final_balance: float
    """balance at the end of the schedule"""
    valid: bool = True
    """Is this a valid schedule?"""
    start: Optional[int] = None
    """The starting step of this schedule"""
    end: Optional[int] = None
    """The step after the last step in this simulation"""
    # these two attributes are not directly related to the schedule but with the scheduling operation that generated it
    needs: List[ProductionNeed] = field(default_factory=list)
    """The products needed but not still in storage needed to complete this schedule."""
    jobs: List[Job] = field(default_factory=list)
    """The jobs that need to be scheduled"""
    failed_contracts: List[Contract] = field(default_factory=list)
    """A list of contracts that failed to be scheduled."""
    ignored_contracts: List[Contract] = field(default_factory=list)
    """A list of contracts ignored for this schedule because they are in the past."""

    def __str__(self):
        fail = str("fail " + "|".join(str(_) for _ in self.failed_contracts)) if len(self.failed_contracts) > 0 else ""
        ignored = str("ignored " + "|".join(str(_) for _ in self.ignored_contracts)) \
            if len(self.ignored_contracts) > 0 else ""
        jobs = str("jobs " + "|".join(str(_) for _ in self.jobs)) if len(self.jobs) > 0 else ""
        updates, needs = '', ''
        if len(self.needs) > 0:
            needs = "needs :" + "\n\t".join(str(_) for _ in self.needs)

        result = f'{"valid" if self.valid else "invalid"} (ends before {self.end}):'
        for x in (fail, ignored, jobs, updates, needs):
            if len(x) > 0:
                result += '\n' + x
        return result

    def combine(self, other: 'ScheduleInfo') -> None:
        self.valid = self.valid and other.valid
        if other.end is not None and self.end is not None:
            self.end = max(self.end, other.end)
        if other.needs is not None:
            self.needs.extend(other.needs)
        if other.jobs is not None:
            self.jobs.extend(other.jobs)
        if other.failed_contracts is not None:
            self.failed_contracts.extend(other.failed_contracts)
        if other.ignored_contracts is not None:
            self.ignored_contracts.extend(other.ignored_contracts)
        self.final_balance = other.final_balance


class Scheduler(ABC):
    """Base class for all schedulers"""

    def __init__(self, manager_id: str, awi: 'SCMLAWI', max_insurance_premium: float, horizon: Optional[int] = None):
        self.horizon = horizon
        self.n_steps = 0
        self.n_lines = 0
        self.simulator: FactorySimulator = None
        self.products: List[Product] = []
        self.processes: List[Process] = []
        self.profiles: List[ManufacturingProfileCompiled] = []
        self.producing: Dict[int, List[ProductManufacturingInfo]] = {}
        self.manager_id = manager_id
        self.awi = awi
        self.max_insurance_premium = max_insurance_premium

    def bookmark(self) -> int:
        """Sets a bookmark to the current location

        Returns:
            bookmark ID
        """
        return self.simulator.bookmark()

    def rollback(self, bookmark_id: int) -> bool:
        """Rolls back to the given bookmark ID

        Args:
            bookmark_id The bookmark ID returned from bookmark

        Remarks:

            - You can only rollback in the reverse order of bookmarks. If the bookmark ID given here is not the one
              at the top of the bookmarks stack, the rollback will fail (return False)

        """
        return self.simulator.rollback(bookmark_id)

    def delete_bookmark(self, bookmark_id: int) -> bool:
        """Commits everything since the bookmark so it cannot be rolled back

        Args:
            bookmark_id The bookmark ID returned from bookmark

        Remarks:

            - You can only rollback in the reverse order of bookmarks. If the bookmark ID given here is not the one
              at the top of the bookmarks stack, the deletion will fail (return False)

        """
        return self.simulator.delete_bookmark(bookmark_id)

    def init(self, simulator: FactorySimulator, products: List[Product], processes: List[Process]
             , profiles: List[ManufacturingProfileCompiled]
             , producing: Dict[int, List[ProductManufacturingInfo]]):
        """Called by the FactoryManager after it is initialized"""
        self.simulator = simulator
        self.n_lines = self.simulator.n_lines
        self.n_steps = self.simulator.n_steps
        self.products = products
        self.processes = processes
        self.producing = producing
        self.profiles = profiles

    def schedule(self, contracts: Collection[Contract] = ()
                 , assume_no_further_negotiations=False
                 , ensure_storage_for: int = 0
                 , start_at: int = 0) -> ScheduleInfo:
        """
        Schedules a set of contracts and returns either the search_for_schedule or None if infeasible

        Args:
            whatever it has scheduled before. If the state is given, it is taken as the initial state for scheduling
            contracts: The contracts to be scheduled
            assume_no_further_negotiations: whether to assume that more negotiations can take place (to secure
            production needs)
            ensure_storage_for: A minimum time to ensure that products are available in storage before contract delivery
            times (sell contracts).
            start_at: The time at which to start scheduling. No jobs will be scheduled before this time.

        Returns:
            `ScheduleInfo` describing the schedulo and any production needs and updates to be carried out.

        """
        # initialize the state of the scheduler with initial knowledge
        # @todo make sure to take start into account. NOw I always schedule from the start
        start, end = 0, -1
        for contract in contracts:
            if contract.agreement['time'] > end:
                end = contract.agreement['time'] + 1
        if self.horizon is not None and end - start > self.horizon:
            end = start + self.horizon
        if end > self.n_steps:
            end = self.n_steps

        # if we have not new contracts, then we are done
        if len(contracts) == 0:
            return ScheduleInfo(valid=True, start=start, end=end
                                , final_balance=self.simulator.final_balance)

        return self.find_schedule(contracts=contracts, start=start, end=end
                                  , assume_no_further_negotiations=assume_no_further_negotiations
                                  , ensure_storage_for=ensure_storage_for
                                  , start_at=start_at)

    @abstractmethod
    def find_schedule(self, contracts: Collection[Contract]
                      , start: int, end: int
                      , assume_no_further_negotiations=False
                      , ensure_storage_for: int = 0
                      , start_at: int = 0) -> ScheduleInfo:
        """
        Schedules a set of contracts and returns either the search_for_schedule or None if infeasible

        Args:
            start:
            end:
            contracts:
            assume_no_further_negotiations:
            ensure_storage_for:
            start_at: The time at which to start scheduling. No jobs will be scheduled before this time.

        Returns:

        """


class GreedyScheduler(Scheduler):
    """Default scheduler used by the DefaultFactoryManager"""

    def __getstate__(self):
        result = self.__dict__.copy()
        if 'fields' in result.keys():
            del result['fields']

    def __setstate__(self, state):
        self.__dict__ = state
        self.fields = [self.total_unit_cost, self.unit_time, self.production_unit_cost, self.input_unit_cost]

    def __init__(self, manager_id: str, awi: 'SCMLAWI', max_insurance_premium: float, horizon: Optional[int] = None
                 , add_catalog_prices=True, strategy: str = 'latest', profile_sorter: str = 'total-cost>time'):
        """

        Args:
            add_catalog_prices: Whether to add total catalog price costs to costs of production
            strategy: How to schedule production. Possible values are earliest, latest, shortest, longest

        Remarks:

            The following `production_strategy` values are supported:

            - earliest: Try to produce things as early as possible. Useful for infinite storage
            - latest: Try to produce things as late as possible. Useful for finite storage
            - shortest: Schedule in the time/line that has the shortest empty slot that is enough for production
            - longest: Schedule in the time/line that has the longest empty slot that is enough for production

            The `profile_sorter` string consists of one or more of the following sections separated by ``>`` characters
            to indicate sorting order. Costs are sorted ascendingly and times descendingly. Costs and times refer to
            unit cost/time (total divided by quantity generated):

            - time, t: profile production time per unit
            - input-cost, ic, icost: Input cost per unit only using catalog prices
            - production-cost, pc, pcost: Production cost as specified in the profile per unit
            - total-cost, tc, tcost: Total cost per unit including input cost


        """
        super().__init__(manager_id=manager_id, horizon=horizon, awi=awi, max_insurance_premium=max_insurance_premium)
        self.add_catalog_prices = add_catalog_prices
        self.strategy = strategy
        self.fields: List[Callable[[ProductManufacturingInfo], float]] = [self.total_unit_cost, self.unit_time
            , self.production_unit_cost, self.input_unit_cost]
        mapper = {'tc': 0, 't': 1, 'pc': 2, 'ic': 3}
        self.field_order: List[int] = []
        sort_fields = profile_sorter.split('>')
        self.producing: Dict[int, List[ProductManufacturingInfo]] = {}
        for field_name in sort_fields:
            if field_name in ('time', 't'):
                self.field_order.append(mapper['t'])
            elif field_name in ('total-cost', 'tc', 'tcost'):
                self.field_order.append(mapper['tc'])
            elif field_name in ('production-cost', 'pc', 'pcost'):
                self.field_order.append(mapper['pc'])
            elif field_name in ('input-cost', 'ic', 'icost'):
                self.field_order.append(mapper['ic'])

    def init(self, simulator: FactorySimulator, products: List[Product], processes: List[Process]
             , profiles: List[ManufacturingProfileCompiled]
             , producing: Dict[int, List[ProductManufacturingInfo]]):
        super().init(simulator=simulator, products=products, processes=processes, producing=producing
                     , profiles=profiles)
        self.producing = {k: sorted(v, key=self._profile_sorter) for k, v in self.producing.items()}

    def _profile_sorter(self, info: ProductManufacturingInfo) -> Any:
        vals = [field(info) for field in self.fields]
        profile = self.profiles[info.profile]
        return tuple([vals[indx] for indx in self.field_order] + [profile.line, profile.process])

    def unit_time(self, info: ProductManufacturingInfo) -> float:
        profile = self.profiles[info.profile]
        return profile.n_steps / info.quantity

    def total_cost(self, info: ProductManufacturingInfo) -> float:
        products = self.products
        profile = self.profiles[info.profile]
        process = self.processes[profile.process]
        production_cost = profile.cost

        def safe(x):
            return 0.0 if x is None else x

        inputs_cost = sum(safe(products[inp.product].catalog_price) * inp.quantity for inp in process.inputs)
        return production_cost + inputs_cost

    def total_unit_cost(self, info: ProductManufacturingInfo) -> float:
        return self.total_cost(info=info) / info.quantity

    def production_cost(self, info: ProductManufacturingInfo) -> float:
        profile = self.profiles[info.profile]
        return profile.cost

    def production_unit_cost(self, info: ProductManufacturingInfo) -> float:
        return self.production_cost(info=info) / info.quantity

    def input_cost(self, info: ProductManufacturingInfo):
        products = self.products
        profile = self.profiles[info.profile]
        process = self.processes[profile.process]

        def safe(x):
            return 0.0 if x is None else x

        return sum(safe(products[inp.product].catalog_price) * inp.quantity for inp in process.inputs)

    def input_unit_cost(self, info: ProductManufacturingInfo) -> float:
        return self.input_cost(info=info) / info.quantity

    # noinspection PyUnusedLocal
    def schedule_contract(self, contract: Contract
                          , assume_no_further_negotiations=False
                          , end: int = None, ensure_storage_for: int = 0
                          , start_at: int = 0) -> ScheduleInfo:
        """
        Schedules this contract if possible and returns information about the resulting schedule

        Args:
            contract:
            assume_no_further_negotiations: If true no further negotiations will be assumed possible
            end:
            ensure_storage_for: The number of steps all needs must be in storage before they are consumed in production
            start_at: No jobs will be scheduled before that time.

        Returns:

        """
        ignore_failures = not assume_no_further_negotiations
        simulator: FactorySimulator = self.simulator
        start = max(simulator.fixed_before, start_at)

        if end is None:
            end = simulator.n_steps
        if contract.agreement is None:
            return ScheduleInfo(end=end, final_balance=self.simulator.balance_at(end - 1))
        agreement: SCMLAgreement
        if isinstance(contract.agreement, dict):
            agreement = SCMLAgreement(**contract.agreement)
        else:
            agreement = contract.agreement  # type: ignore
        t = agreement['time']
        if t < start:
            return ScheduleInfo(end=end, final_balance=INVALID_UTILITY, valid=False, ignored_contracts=[contract])
        q, u = int(agreement['quantity']), agreement['unit_price']
        p = u * q
        pid: int = contract.annotation['cfp'].product
        if contract.annotation['buyer'] == self.manager_id:
            # I am a buyer
            # We do not ignore money shortage for buying. This means that the agent will not buy if the money it needs
            # may partially come from a sell contract that is not considered yet
            if not simulator.buy(product=pid, quantity=q, price=p, t=t, ignore_space_shortage=ignore_failures
                , ignore_money_shortage=ignore_failures):
                return ScheduleInfo(end=end, valid=False, failed_contracts=[contract], final_balance=INVALID_UTILITY)
            if p <= 0:
                return ScheduleInfo(valid=True, end=end, final_balance=self.simulator.balance_at(end - 1))
            insurance = self.awi.evaluate_insurance(contract=contract, t=self.awi.current_step)
            if insurance is not None and insurance / p < self.max_insurance_premium:
                # if it is not possible to buy the insurance, the factory manager will not try to buy it. This is still
                # a valid schedule
                simulator.pay(insurance, t=t)
            return ScheduleInfo(valid=True, end=end, final_balance=self.simulator.balance_at(end - 1))
        elif contract.annotation['seller'] == self.manager_id:
            # I am a seller

            # if enough is available in storage and not reserved, just sell it
            q_needed = q - simulator.available_storage_at(t)[pid]
            if q_needed <= 0:
                if simulator.sell(product=pid, quantity=q, price=p, t=t, ignore_money_shortage=ignore_failures
                    , ignore_inventory_shortage=ignore_failures):
                    return ScheduleInfo(end=end, final_balance=self.simulator.balance_at(end - 1))
                else:
                    return ScheduleInfo(end=end, valid=False, failed_contracts=[contract],
                                        final_balance=INVALID_UTILITY)
            jobs: List[Job] = []
            needs: List[ProductionNeed] = []

            with transaction(simulator) as bookmark:
                some_production = True
                while q_needed > 0 and some_production:
                    some_production = False
                    # I need now to schedule the production needed and calculate all required input products
                    for info in self.producing[pid]:
                        # find if it is possible to use the current process for producing the product
                        profile = self.profiles[info.profile]
                        line, process_index, profile_index = profile.line, profile.process, info.profile
                        q_produced, t_production = info.quantity, info.step
                        current_schedule = simulator.line_schedules_to(t - ensure_storage_for - 1)[line][start:]
                        if len(current_schedule) < t_production:
                            continue
                        locs = zero_runs((current_schedule != NO_PRODUCTION).astype(int))
                        lengths = locs[:, 1] - locs[:, 0]
                        indices = np.array(range(len(lengths)))
                        indices = indices[lengths >= t_production]
                        if len(indices) < 1:
                            continue
                        lengths, locs = lengths[indices], locs[indices]
                        if self.strategy == 'earliest':
                            loc = locs[0, :]
                        elif self.strategy == 'latest':
                            loc = locs[-1, :] - 1
                            loc[0] = loc[1] - t_production + 1
                        elif self.strategy == 'shortest':
                            sorted_lengths = sorted(zip(range(len(lengths)), lengths), key=lambda x: x[1])
                            loc = locs[sorted_lengths[0][0], :]
                        elif self.strategy == 'longest':
                            sorted_lengths = sorted(zip(range(len(lengths)), lengths), key=lambda x: x[1], reverse=True)
                            loc = locs[sorted_lengths[0][0], :]
                        else:
                            raise ValueError(f'Unknown production strategy {self.strategy}')
                        ptime = loc[0] + start
                        job = Job(line=line, action='run', time=ptime, profile=profile_index
                                  , contract=contract, override=False)
                        if not simulator.schedule(job, override=False, ignore_inventory_shortage=ignore_failures
                            , ignore_money_shortage=ignore_failures
                            , ignore_space_shortage=ignore_failures):
                            continue  # should never hit this
                        jobs.append(job)
                        # find the needs
                        new_needs = []
                        process = self.processes[process_index]
                        length = profile.n_steps
                        for i in process.inputs:
                            pind, quantity = i.product, i.quantity
                            # I need the input to be available the step before production
                            step = ptime + int(math.floor(i.step * length)) - 1
                            if step < 0:
                                break
                            available = max(0, self.simulator.available_storage_at(step)[pind] - quantity)
                            if available >= quantity:
                                instore, tobuy = quantity, 0
                            else:
                                instore, tobuy = available, quantity - available
                            if tobuy > 0 or instore > 0:
                                if step < start:
                                    break
                                needs.append(ProductionNeed(product=pind, needed_for=contract
                                                            , quantity_in_storage=instore, quantity_to_buy=tobuy
                                                            , step=step))
                        else:  # all inputs can be secured in time
                            # @todo consider stopping production after the product is available (+ ensure_storage_for) if needed
                            q_needed -= q_produced
                            some_production = True
                            break
                if q_needed <= 0:
                    # add the effect of selling
                    if not simulator.sell(product=pid, quantity=q, price=p, t=t, ignore_money_shortage=ignore_failures
                        , ignore_inventory_shortage=ignore_failures):
                        simulator.rollback(bookmark)
                        return ScheduleInfo(end=end, valid=False, failed_contracts=[contract],
                                            final_balance=INVALID_UTILITY)

                    # add the effect of buying raw materials
                    for need in needs:
                        product_index = need.product
                        product = self.products[product_index]
                        catalog_price = product.catalog_price
                        if catalog_price == 0 or need.quantity_to_buy <= 0:
                            continue
                        price = need.quantity_to_buy * catalog_price
                        simulator.pay(price, t=need.step)

                    # create schedule
                    schedule = ScheduleInfo(jobs=jobs, end=end, needs=needs, failed_contracts=[]
                                            , final_balance=self.simulator.balance_at(end - 1))
                    return schedule
                simulator.rollback(bookmark)
                return ScheduleInfo(valid=False, failed_contracts=[contract], end=end
                                    , final_balance=self.simulator.balance_at(end - 1))
        raise ValueError(f'{self.manager_id} Not a seller of a buyer in Contract: {contract} with '
                         f'annotation: {contract.annotation}')

    def schedule_contracts(self, contracts: Collection[Contract], end: int = None
                           , assume_no_further_negotiations=False
                           , ensure_storage_for: int = 0
                           , start_at: int = 0) -> ScheduleInfo:
        """
        Schedules a set of contracts and returns the `ScheduleInfo`.

        Args:
            contracts: Contracts to schedule
            assume_no_further_negotiations: If true, no further negotiations will be assumed to be possible
            end: The end of the simulation for the schedule (exclusive)
            ensure_storage_for: Ensure that the outcome will be at the storage for at least this time
            start_at: The timestep at which to start scheduling

        Returns:

            ScheduleInfo giving the schedule after these contracts is included. `valid` member can be used to check
            whether this is a valid contract

        """
        simulator = self.simulator
        if end is None:
            end = simulator.n_steps
        result = ScheduleInfo(valid=True, end=end, final_balance=self.simulator.final_balance)
        contracts = sorted(contracts, key=lambda x: x.agreement['time'])
        for contract in contracts:
            new_schedule = self.schedule_contract(contract, end=end, ensure_storage_for=ensure_storage_for
                                                  , assume_no_further_negotiations=assume_no_further_negotiations
                                                  , start_at=start_at)
            result.combine(new_schedule)
            if new_schedule.valid:
                result.final_balance = self.simulator.final_balance
            else:
                result.final_balance = INVALID_UTILITY
        return result

    def find_schedule(self, contracts: Collection[Contract]
                      , start: int, end: int
                      , assume_no_further_negotiations=False
                      , ensure_storage_for: int = 0
                      , start_at: int = 0):

        # Now, schedule the contracts
        schedule = self.schedule_contracts(contracts=contracts, end=end, ensure_storage_for=ensure_storage_for
                                           , assume_no_further_negotiations=assume_no_further_negotiations
                                           , start_at=start_at)

        # Mark the schedule as invalid if it has any production needs and we assume_no_further_negotiations
        if assume_no_further_negotiations and schedule.needs is not None and len(schedule.needs) > 0:
            schedule.valid = False
            return schedule

        return schedule
